from __future__ import annotations

import sys
from typing import Any

from docutils.parsers.rst import Directive

import contourpy

sys.path.insert(0, ".")
from table import Table


class NameSupportsType(Directive):
    required_arguments = 1

    def run(self) -> list[Any]:
        if self.arguments[0] not in ("LineType", "FillType"):
            raise ValueError(f"Do not recognise argument {self.arguments[0]}")

        type_name = self.arguments[0]
        lowercase_name = type_name.replace(r"Type", "_type").lower()
        default_func_name = f"default_{lowercase_name}"
        supports_func_name = f"supports_{lowercase_name}"
        type_enum = getattr(contourpy, type_name)

        names = list(contourpy._class_lookup)
        classes = list(contourpy._class_lookup.values())
        default_types = [getattr(cls, default_func_name).name for cls in classes]

        table = Table(1 + len(names))
        table.add_header([type_name, *names])
        for name, enum in dict(type_enum.__members__).items():
            row = [name]
            for cls, default_type in zip(classes, default_types):
                func = getattr(cls, supports_func_name)
                if func(enum):
                    cell = "Yes"
                    if name == default_type:
                        cell += ", default"
                    row.append(cell)
                else:
                    row.append("")
            table.add_row(row)

        return [table.get()]


def setup(app: Any) -> dict[str, bool]:
    app.add_directive("name_supports_type", NameSupportsType)
    return {"parallel_read_safe": True, "parallel_write_safe": True}



from __future__ import annotations

import pytest

from contourpy import (
    ContourGenerator, FillType, LineType, Mpl2005ContourGenerator, Mpl2014ContourGenerator,
    SerialContourGenerator, ThreadedContourGenerator,
)

from . import util_test

_lookup: dict[str, type[ContourGenerator]] = {
    "Mpl2005ContourGenerator": Mpl2005ContourGenerator,
    "Mpl2014ContourGenerator": Mpl2014ContourGenerator,
    "SerialContourGenerator": SerialContourGenerator,
    "ThreadedContourGenerator": ThreadedContourGenerator,
}


def get_class_from_name(class_name: str) -> type[ContourGenerator]:
    return _lookup[class_name]


@pytest.mark.parametrize("class_name", util_test.all_class_names())
def test_default_fill_type(class_name: str) -> None:
    cls = get_class_from_name(class_name)
    default = cls.default_fill_type
    assert isinstance(default, FillType)
    if class_name in ("Mpl2005ContourGenerator", "Mpl2014ContourGenerator"):
        expect = FillType.OuterCode
    else:
        expect = FillType.OuterOffset
    assert default == expect


@pytest.mark.parametrize("class_name", util_test.all_class_names())
def test_default_line_type(class_name: str) -> None:
    cls = get_class_from_name(class_name)
    default = cls.default_line_type
    assert isinstance(default, LineType)
    if class_name in ("Mpl2005ContourGenerator", "Mpl2014ContourGenerator"):
        expect = LineType.SeparateCode
    else:
        expect = LineType.Separate
    assert default == expect


@pytest.mark.parametrize("class_name", util_test.all_class_names())
def test_has_lines_and_filled(class_name: str) -> None:
    cls = get_class_from_name(class_name)
    for func_name in ["create_contour", "create_filled_contour", "filled", "lines"]:
        assert hasattr(cls, func_name)
        assert callable(getattr(cls, func_name))


@pytest.mark.parametrize("class_name", util_test.all_class_names())
def test_supports_corner_mask(class_name: str) -> None:
    cls = get_class_from_name(class_name)
    supports = cls.supports_corner_mask()
    assert isinstance(supports, bool)
    expect = class_name != "Mpl2005ContourGenerator"
    assert supports == expect


@pytest.mark.parametrize("class_name", util_test.all_class_names())
def test_supports_fill_type(class_name: str) -> None:
    cls = get_class_from_name(class_name)
    supports = cls.supports_fill_type(FillType.OuterCode)
    assert isinstance(supports, bool)
    expect = True
    assert supports == expect
    supports = cls.supports_fill_type(FillType.ChunkCombinedOffsetOffset)
    assert isinstance(supports, bool)
    expect = class_name not in ("Mpl2005ContourGenerator", "Mpl2014ContourGenerator")
    assert supports == expect


@pytest.mark.parametrize("class_name", util_test.all_class_names())
def test_supports_line_type(class_name: str) -> None:
    cls = get_class_from_name(class_name)
    supports = cls.supports_line_type(LineType.SeparateCode)
    assert isinstance(supports, bool)
    expect = True
    assert supports == expect
    supports = cls.supports_line_type(LineType.ChunkCombinedOffset)
    assert isinstance(supports, bool)
    expect = class_name not in ("Mpl2005ContourGenerator", "Mpl2014ContourGenerator")
    assert supports == expect


@pytest.mark.parametrize("class_name", util_test.all_class_names())
def test_supports_quad_as_tri(class_name: str) -> None:
    cls = get_class_from_name(class_name)
    supports = cls.supports_quad_as_tri()
    assert isinstance(supports, bool)
    expect = class_name not in ("Mpl2005ContourGenerator", "Mpl2014ContourGenerator")
    assert supports == expect


@pytest.mark.parametrize("class_name", util_test.all_class_names())
def test_supports_threads(class_name: str) -> None:
    cls = get_class_from_name(class_name)
    supports = cls.supports_threads()
    assert isinstance(supports, bool)
    expect = class_name == "ThreadedContourGenerator"
    assert supports == expect


@pytest.mark.parametrize("class_name", util_test.all_class_names())
def test_supports_z_interp(class_name: str) -> None:
    cls = get_class_from_name(class_name)
    supports = cls.supports_z_interp()
    assert isinstance(supports, bool)
    expect = class_name not in ("Mpl2005ContourGenerator", "Mpl2014ContourGenerator")
    assert supports == expect



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import numpy as np
from numpy.testing import assert_allclose, assert_array_equal
import pytest

from contourpy import LineType, contour_generator
from contourpy.util.data import random, simple

from . import util_test

if TYPE_CHECKING:
    import contourpy._contourpy as cpy


@pytest.fixture
def xy_2x2() -> tuple[cpy.CoordinateArray, ...]:
    x, y = np.meshgrid([0.0, 1.0], [0.0, 1.0])
    return x, y


@pytest.fixture
def xy_3x3() -> tuple[cpy.CoordinateArray, ...]:
    x, y = np.meshgrid([0.0, 0.5, 1.0], [0.0, 0.5, 1.0])
    return x, y


@pytest.fixture
def one_loop_one_strip() -> tuple[cpy.PointArray, ...]:
    x, y = np.meshgrid([0., 1., 2., 3.], [0., 1., 2.])
    z = np.array([[1.5, 1.5, 0.9, 0.0],
                  [1.5, 2.8, 0.4, 0.8],
                  [0.0, 0.0, 0.8, 6.0]])
    return x, y, z


@pytest.fixture
def xyz_chunk_test() -> tuple[cpy.CoordinateArray, ...]:
    x, y = np.meshgrid(np.arange(5), np.arange(5))
    z = 0.5*np.abs(y - 2) + 0.1*(x - 2)
    return x, y, z


@pytest.mark.parametrize("name", util_test.all_names())
@pytest.mark.parametrize("zlevel", [-1e-10, 1.0+1e10, np.nan, -np.nan, np.inf, -np.inf])
def test_level_outside(xy_2x2: tuple[cpy.CoordinateArray, ...], name: str, zlevel: float) -> None:
    x, y = xy_2x2
    z = x
    cont_gen = contour_generator(x, y, z, name=name, line_type=LineType.SeparateCode)
    lines = cont_gen.lines(zlevel)
    if TYPE_CHECKING:
        lines = cast(cpy.LineReturn_SeparateCode, lines)
    assert isinstance(lines, tuple) and len(lines) == 2
    points, codes = lines
    assert isinstance(points, list) and len(points) == 0
    assert isinstance(codes, list) and len(codes) == 0


@pytest.mark.parametrize("name", util_test.all_names())
def test_w_to_e(xy_2x2: tuple[cpy.CoordinateArray, ...], name: str) -> None:
    x, y = xy_2x2
    z = y.copy()
    cont_gen = contour_generator(x, y, z, name=name, line_type=LineType.SeparateCode)
    lines = cont_gen.lines(0.5)
    if TYPE_CHECKING:
        lines = cast(cpy.LineReturn_SeparateCode, lines)
    assert isinstance(lines, tuple) and len(lines) == 2
    points, codes = lines
    assert isinstance(points, list) and len(points) == 1
    assert isinstance(codes, list) and len(codes) == 1
    assert_allclose(points[0], [[0.0, 0.5], [1.0, 0.5]])
    assert_array_equal(codes[0], [1, 2])


@pytest.mark.parametrize("name", util_test.all_names())
def test_e_to_w(xy_2x2: tuple[cpy.CoordinateArray, ...], name: str) -> None:
    x, y = xy_2x2
    z = 1.0 - y.copy()
    cont_gen = contour_generator(x, y, z, name=name, line_type=LineType.SeparateCode)
    lines = cont_gen.lines(0.5)
    if TYPE_CHECKING:
        lines = cast(cpy.LineReturn_SeparateCode, lines)
    assert isinstance(lines, tuple) and len(lines) == 2
    points, codes = lines
    assert isinstance(points, list) and len(points) == 1
    assert isinstance(codes, list) and len(codes) == 1
    if name == "mpl2005":    # Line directions are not consistent.
        assert_allclose(points[0], [[0.0, 0.5], [1.0, 0.5]])
    else:
        assert_allclose(points[0], [[1.0, 0.5], [0.0, 0.5]])
    assert_array_equal(codes[0], [1, 2])


@pytest.mark.parametrize("name", util_test.all_names())
def test_loop(xy_3x3: tuple[cpy.CoordinateArray, ...], name: str) -> None:
    x, y = xy_3x3
    z = np.zeros_like(x)
    z[1, 1] = 1.0
    cont_gen = contour_generator(x, y, z, name=name, line_type=LineType.SeparateCode)
    lines = cont_gen.lines(0.5)
    if TYPE_CHECKING:
        lines = cast(cpy.LineReturn_SeparateCode, lines)
    assert isinstance(lines, tuple) and len(lines) == 2
    points, codes = lines
    assert isinstance(points, list) and len(points) == 1
    assert isinstance(codes, list) and len(codes) == 1
    line = points[0]
    assert line.shape == (5, 2)
    assert_allclose(line[0], line[-1])
    assert_array_equal(codes[0], [1, 2, 2, 2, 79])


@pytest.mark.image
@pytest.mark.parametrize("name, line_type", util_test.all_names_and_line_types())
def test_lines_simple(name: str, line_type: LineType) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = simple((30, 40))
    cont_gen = contour_generator(x, y, z, name=name, line_type=line_type)
    levels = np.arange(-1.0, 1.01, 0.1)

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (1, 1)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)):
        renderer.lines(cont_gen.lines(levels[i]), line_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "lines_simple.png", f"{name}_{line_type}")


@pytest.mark.image
@pytest.mark.parametrize("name, line_type", util_test.all_names_and_line_types())
def test_lines_simple_chunk(name: str, line_type: LineType) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = simple((30, 40))
    cont_gen = contour_generator(
        x, y, z, name=name, line_type=line_type, chunk_size=2, thread_count=1,
    )
    levels = np.arange(-1.0, 1.01, 0.1)

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)):
        renderer.lines(cont_gen.lines(levels[i]), line_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "lines_simple_chunk.png", f"{name}_{line_type}")


@pytest.mark.image
@pytest.mark.threads
@pytest.mark.parametrize("line_type", LineType.__members__.values())
@pytest.mark.parametrize("thread_count", util_test.thread_counts())
def test_lines_simple_chunk_threads(line_type: LineType, thread_count: int) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = simple((30, 40))
    cont_gen = contour_generator(
        x, y, z, name="threaded", line_type=line_type, chunk_size=2, thread_count=thread_count,
    )
    levels = np.arange(-1.0, 1.01, 0.1)

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == thread_count

    renderer = MplTestRenderer()
    for i in range(len(levels)):
        renderer.lines(cont_gen.lines(levels[i]), line_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "lines_simple_chunk.png", f"{line_type}_{thread_count}")


@pytest.mark.image
@pytest.mark.parametrize("name, line_type", util_test.all_names_and_line_types())
def test_lines_simple_no_corner_mask(name: str, line_type: LineType) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = simple((30, 40), want_mask=True)
    cont_gen = contour_generator(x, y, z, name=name, line_type=line_type, corner_mask=False)
    levels = np.arange(-1.0, 1.01, 0.1)

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (1, 1)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)):
        renderer.lines(cont_gen.lines(levels[i]), line_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "lines_simple_no_corner_mask.png", f"{name}_{line_type}")


@pytest.mark.image
@pytest.mark.parametrize("name, line_type", util_test.all_names_and_line_types())
def test_lines_simple_no_corner_mask_chunk(name: str, line_type: LineType) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = simple((30, 40), want_mask=True)
    cont_gen = contour_generator(
        x, y, z, name=name, line_type=line_type, corner_mask=False, chunk_size=2, thread_count=1,
    )
    levels = np.arange(-1.0, 1.01, 0.1)

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)):
        renderer.lines(cont_gen.lines(levels[i]), line_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "lines_simple_no_corner_mask_chunk.png", f"{name}_{line_type}")


@pytest.mark.image
@pytest.mark.threads
@pytest.mark.parametrize("line_type", LineType.__members__.values())
@pytest.mark.parametrize("thread_count", util_test.thread_counts())
def test_lines_simple_no_corner_mask_chunk_threads(line_type: LineType, thread_count: int) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = simple((30, 40), want_mask=True)
    cont_gen = contour_generator(
        x, y, z, name="threaded", line_type=line_type, corner_mask=False, chunk_size=2,
        thread_count=thread_count,
    )
    levels = np.arange(-1.0, 1.01, 0.1)

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == thread_count

    renderer = MplTestRenderer()
    for i in range(len(levels)):
        renderer.lines(cont_gen.lines(levels[i]), line_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(
        image_buffer, "lines_simple_no_corner_mask_chunk.png", f"{line_type}_{thread_count}",
    )


@pytest.mark.image
@pytest.mark.parametrize("name", util_test.corner_mask_names())
def test_lines_simple_corner_mask(name: str) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = simple((30, 40), want_mask=True)
    line_type = LineType.SeparateCode
    cont_gen = contour_generator(x, y, z, name=name, line_type=line_type, corner_mask=True)
    levels = np.arange(-1.0, 1.01, 0.1)

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (1, 1)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)):
        renderer.lines(cont_gen.lines(levels[i]), line_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "lines_simple_corner_mask.png", f"{name}_{line_type}")


@pytest.mark.image
@pytest.mark.parametrize("name", util_test.corner_mask_names())
def test_lines_simple_corner_mask_chunk(name: str) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = simple((30, 40), want_mask=True)
    line_type = LineType.SeparateCode
    cont_gen = contour_generator(
        x, y, z, name=name, line_type=line_type, corner_mask=True, chunk_size=2, thread_count=1,
    )
    levels = np.arange(-1.0, 1.01, 0.1)

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)):
        renderer.lines(cont_gen.lines(levels[i]), line_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "lines_simple_corner_mask_chunk.png", f"{name}_{line_type}")


@pytest.mark.image
@pytest.mark.threads
@pytest.mark.parametrize("line_type", LineType.__members__.values())
@pytest.mark.parametrize("thread_count", util_test.thread_counts())
def test_lines_simple_corner_mask_chunk_threads(line_type: LineType, thread_count: int) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = simple((30, 40), want_mask=True)
    cont_gen = contour_generator(
        x, y, z, name="threaded", line_type=line_type, corner_mask=True, chunk_size=2,
        thread_count=thread_count,
    )
    levels = np.arange(-1.0, 1.01, 0.1)

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == thread_count

    renderer = MplTestRenderer()
    for i in range(len(levels)):
        renderer.lines(cont_gen.lines(levels[i]), line_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(
        image_buffer, "lines_simple_corner_mask_chunk.png", f"{line_type}_{thread_count}",
    )


@pytest.mark.image
@pytest.mark.parametrize("name", util_test.quad_as_tri_names())
def test_lines_simple_quad_as_tri(name: str) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = simple((30, 40))
    cont_gen = contour_generator(x, y, z, name=name, quad_as_tri=True)
    levels = np.arange(-1.0, 1.01, 0.1)

    assert cont_gen.chunk_count == (1, 1)
    assert cont_gen.thread_count == 1
    assert cont_gen.quad_as_tri

    line_type = cont_gen.line_type
    renderer = MplTestRenderer()
    for i in range(len(levels)):
        renderer.lines(cont_gen.lines(levels[i]), line_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "lines_simple_quad_as_tri.png", f"{name}")


@pytest.mark.image
@pytest.mark.parametrize("name, line_type", util_test.all_names_and_line_types())
def test_lines_random(name: str, line_type: LineType) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = random((30, 40), mask_fraction=0.0)
    cont_gen = contour_generator(x, y, z, name=name, line_type=line_type)
    levels = np.arange(0.0, 1.01, 0.2)

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (1, 1)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)):
        renderer.lines(cont_gen.lines(levels[i]), line_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "lines_random.png", f"{name}_{line_type}", max_threshold=103)


@pytest.mark.image
@pytest.mark.parametrize("name, line_type", util_test.all_names_and_line_types())
def test_lines_random_chunk(name: str, line_type: LineType) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = random((30, 40), mask_fraction=0.0)
    cont_gen = contour_generator(
        x, y, z, name=name, line_type=line_type, chunk_size=2, thread_count=1,
    )
    levels = np.arange(0.0, 1.01, 0.2)

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)):
        renderer.lines(cont_gen.lines(levels[i]), line_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    max_threshold = None
    mean_threshold = None
    if name == "mpl2005":
        max_threshold = 126
        mean_threshold = 0.11

    compare_images(
        image_buffer, "lines_random_chunk.png", f"{name}_{line_type}",
        max_threshold=max_threshold, mean_threshold=mean_threshold,
    )


@pytest.mark.image
@pytest.mark.threads
@pytest.mark.parametrize("line_type", LineType.__members__.values())
@pytest.mark.parametrize("thread_count", util_test.thread_counts())
def test_lines_random_chunk_threads(line_type: LineType, thread_count: int) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = random((30, 40), mask_fraction=0.0)
    cont_gen = contour_generator(
        x, y, z, name="threaded", line_type=line_type, chunk_size=2, thread_count=thread_count,
    )
    levels = np.arange(0.0, 1.01, 0.2)

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == thread_count

    renderer = MplTestRenderer()
    for i in range(len(levels)):
        renderer.lines(cont_gen.lines(levels[i]), line_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(
        image_buffer, "lines_random_chunk.png", f"{line_type}_{thread_count}",
    )


@pytest.mark.image
@pytest.mark.parametrize("name, line_type", util_test.all_names_and_line_types())
def test_lines_random_no_corner_mask(name: str, line_type: LineType) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = random((30, 40), mask_fraction=0.05)
    cont_gen = contour_generator(x, y, z, name=name, line_type=line_type, corner_mask=False)
    levels = np.arange(0.0, 1.01, 0.2)

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (1, 1)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)):
        renderer.lines(cont_gen.lines(levels[i]), line_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "lines_random_no_corner_mask.png", f"{name}_{line_type}")


@pytest.mark.image
@pytest.mark.parametrize("name, line_type", util_test.all_names_and_line_types())
def test_lines_random_no_corner_mask_chunk(name: str, line_type: LineType) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    if name in ("mpl2005"):
        pytest.skip()  # mpl2005 does not support chunks for lines.

    x, y, z = random((30, 40), mask_fraction=0.05)
    cont_gen = contour_generator(
        x, y, z, name=name, line_type=line_type, corner_mask=False, chunk_size=2, thread_count=1,
    )
    levels = np.arange(0.0, 1.01, 0.2)

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)):
        renderer.lines(cont_gen.lines(levels[i]), line_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "lines_random_no_corner_mask_chunk.png", f"{name}_{line_type}")


@pytest.mark.image
@pytest.mark.threads
@pytest.mark.parametrize("line_type", LineType.__members__.values())
@pytest.mark.parametrize("thread_count", util_test.thread_counts())
def test_lines_random_no_corner_mask_chunk_threads(line_type: LineType, thread_count: int) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = random((30, 40), mask_fraction=0.05)
    cont_gen = contour_generator(
        x, y, z, name="threaded", line_type=line_type, corner_mask=False, chunk_size=2,
        thread_count=thread_count,
    )
    levels = np.arange(0.0, 1.01, 0.2)

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == thread_count

    renderer = MplTestRenderer()
    for i in range(len(levels)):
        renderer.lines(cont_gen.lines(levels[i]), line_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(
        image_buffer, "lines_random_no_corner_mask_chunk.png", f"{line_type}_{thread_count}",
    )


@pytest.mark.image
@pytest.mark.parametrize("name", util_test.corner_mask_names())
def test_lines_random_corner_mask(name: str) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = random((30, 40), mask_fraction=0.05)
    line_type = LineType.SeparateCode
    cont_gen = contour_generator(x, y, z, name=name, corner_mask=True, line_type=line_type)
    levels = np.arange(0.0, 1.01, 0.2)

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (1, 1)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)):
        renderer.lines(cont_gen.lines(levels[i]), line_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "lines_random_corner_mask.png", f"{name}_{line_type}")


@pytest.mark.image
@pytest.mark.parametrize("name", util_test.corner_mask_names())
def test_lines_random_corner_mask_chunk(name: str) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = random((30, 40), mask_fraction=0.05)
    line_type = LineType.SeparateCode
    cont_gen = contour_generator(
        x, y, z, name=name, corner_mask=True, line_type=line_type, chunk_size=2, thread_count=1,
    )
    levels = np.arange(0.0, 1.01, 0.2)

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)):
        renderer.lines(cont_gen.lines(levels[i]), line_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "lines_random_corner_mask_chunk.png", f"{name}_{line_type}")


@pytest.mark.image
@pytest.mark.threads
@pytest.mark.parametrize("line_type", LineType.__members__.values())
@pytest.mark.parametrize("thread_count", util_test.thread_counts())
def test_lines_random_corner_mask_chunk_threads(line_type: LineType, thread_count: int) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = random((30, 40), mask_fraction=0.05)
    cont_gen = contour_generator(
        x, y, z, name="threaded", corner_mask=True, line_type=line_type, chunk_size=2,
        thread_count=thread_count,
    )
    levels = np.arange(0.0, 1.01, 0.2)

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == thread_count

    renderer = MplTestRenderer()
    for i in range(len(levels)):
        renderer.lines(cont_gen.lines(levels[i]), line_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(
        image_buffer, "lines_random_corner_mask_chunk.png", f"{line_type}_{thread_count}",
    )


@pytest.mark.image
@pytest.mark.parametrize("name", util_test.quad_as_tri_names())
def test_lines_random_quad_as_tri(name: str) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = random((30, 40), mask_fraction=0.0)
    cont_gen = contour_generator(x, y, z, name=name, quad_as_tri=True)
    levels = np.arange(0.0, 1.01, 0.2)

    assert cont_gen.chunk_count == (1, 1)
    assert cont_gen.thread_count == 1
    assert cont_gen.quad_as_tri

    line_type = cont_gen.line_type
    renderer = MplTestRenderer()
    for i in range(len(levels)):
        renderer.lines(cont_gen.lines(levels[i]), line_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "lines_random_quad_as_tri.png", f"{name}")


@pytest.mark.parametrize("line_type", LineType.__members__.values())
@pytest.mark.parametrize("name", ["serial", "threaded"])
def test_return_by_line_type(
    one_loop_one_strip: tuple[cpy.PointArray, ...],
    name: str,
    line_type: LineType,
) -> None:
    x, y, z = one_loop_one_strip
    cont_gen = contour_generator(x, y, z, name=name, line_type=line_type)

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (1, 1)
    assert cont_gen.thread_count == 1

    lines = cont_gen.lines(2.0)

    util_test.assert_lines(lines, line_type)

    if line_type == LineType.Separate:
        if TYPE_CHECKING:
            lines = cast(cpy.LineReturn_Separate, lines)
        points = lines
        assert len(points) == 2
        assert points[0].shape == (5, 2)
        assert points[1].shape == (2, 2)
    elif line_type == LineType.SeparateCode:
        if TYPE_CHECKING:
            lines = cast(cpy.LineReturn_SeparateCode, lines)
        points, codes = lines
        assert len(points) == 2
        assert points[0].shape == (5, 2)
        assert points[1].shape == (2, 2)
        assert_array_equal(codes[0], [1, 2, 2, 2, 79])
        assert_array_equal(codes[1], [1, 2])
    elif line_type == LineType.ChunkCombinedCode:
        if TYPE_CHECKING:
            lines = cast(cpy.LineReturn_ChunkCombinedCode, lines)
        assert len(lines[0]) == 1  # Single chunk.
        points_or_none, codes_or_none = lines[0][0], lines[1][0]
        assert points_or_none is not None
        assert codes_or_none is not None
        assert points_or_none.shape == (7, 2)
        assert_array_equal(codes_or_none, [1, 2, 2, 2, 79, 1, 2])
    elif line_type == LineType.ChunkCombinedOffset:
        if TYPE_CHECKING:
            lines = cast(cpy.LineReturn_ChunkCombinedOffset, lines)
        assert len(lines[0]) == 1  # Single chunk.
        points_or_none, offsets_or_none = lines[0][0], lines[1][0]
        assert points_or_none is not None
        assert offsets_or_none is not None
        assert points_or_none.shape == (7, 2)
        assert_array_equal(offsets_or_none, [0, 5, 7])
    else:
        raise RuntimeError(f"Unexpected line_type {line_type}")


@pytest.mark.threads
@pytest.mark.parametrize("line_type", LineType.__members__.values())
@pytest.mark.parametrize("name, thread_count",
                         [("serial", 1), ("threaded", 1), ("threaded", 2)])
def test_return_by_line_type_chunk(
    xyz_chunk_test: tuple[cpy.CoordinateArray, ...],
    name: str,
    thread_count: int,
    line_type: LineType,
) -> None:
    x, y, z = xyz_chunk_test
    cont_gen = contour_generator(
        x, y, z, name=name, line_type=line_type, chunk_count=2, thread_count=thread_count,
    )

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (2, 2)
    assert cont_gen.chunk_size == (2, 2)
    assert cont_gen.thread_count == thread_count

    lines = cont_gen.lines(0.45)

    util_test.assert_lines(lines, line_type)

    # Expected points by chunk.
    expected: list[cpy.PointArray] = [
        np.asarray([[2.0, 1.1], [1.5, 1.0], [1.0, 0.9], [0.0, 0.7]]),
        np.asarray([[4.0, 1.5], [3.0, 1.3], [2.0, 1.1]]),
        np.asarray([[0.0, 3.3], [1.0, 3.1], [1.5, 3.0], [2.0, 2.9]]),
        np.asarray([[2.0, 2.9], [3.0, 2.7], [4.0, 2.5]]),
    ]

    if line_type == LineType.Separate:
        if TYPE_CHECKING:
            lines = cast(cpy.LineReturn_Separate, lines)
        assert len(lines) == 4
        if name == "threaded" and cont_gen.thread_count > 1:
            # Lines may be in any order so sort lines and expected.
            lines = util_test.sort_by_first_xy(lines)
            expected = util_test.sort_by_first_xy(expected)
        for chunk in range(4):
            assert_allclose(lines[chunk], expected[chunk])
    elif line_type == LineType.SeparateCode:
        if TYPE_CHECKING:
            lines = cast(cpy.LineReturn_SeparateCode, lines)
        assert len(lines[0]) == 4
        if name == "threaded" and cont_gen.thread_count > 1:
            # Lines may be in any order so sort lines and expected.
            lines = util_test.sort_by_first_xy(lines[0], lines[1])
            expected = util_test.sort_by_first_xy(expected)
        for chunk in range(4):
            assert_allclose(lines[0][chunk], expected[chunk])
            assert_array_equal(lines[1][chunk], [1] + [2]*(len(expected[chunk])-1))
    elif line_type == LineType.ChunkCombinedCode:
        if TYPE_CHECKING:
            lines = cast(cpy.LineReturn_ChunkCombinedCode, lines)
        assert len(lines[0]) == 4
        for chunk in range(4):
            points_or_none, codes_or_none = lines[0][chunk], lines[1][chunk]
            assert points_or_none is not None
            assert codes_or_none is not None
            assert_allclose(points_or_none, expected[chunk])
            assert_array_equal(codes_or_none, [1] + [2]*(len(expected[chunk])-1))
    elif line_type == LineType.ChunkCombinedOffset:
        if TYPE_CHECKING:
            lines = cast(cpy.LineReturn_ChunkCombinedOffset, lines)
        assert len(lines[0]) == 4
        for chunk in range(4):
            points_or_none, offsets_or_none = lines[0][chunk], lines[1][chunk]
            assert points_or_none is not None
            assert offsets_or_none is not None
            assert_allclose(points_or_none, expected[chunk])
            assert_array_equal(offsets_or_none, [0, len(expected[chunk])])
    else:
        raise RuntimeError(f"Unexpected line_type {line_type}")


@pytest.mark.parametrize("name, line_type", util_test.all_names_and_line_types())
@pytest.mark.parametrize("corner_mask", [None, False, True])
def test_lines_random_big(name: str, line_type: LineType, corner_mask: bool) -> None:
    if corner_mask and name in ["mpl2005", "mpl2014"]:
        pytest.skip()

    x, y, z = random((1000, 1000), mask_fraction=0.0 if corner_mask is None else 0.05)
    cont_gen = contour_generator(x, y, z, name=name, corner_mask=corner_mask, line_type=line_type)
    levels = np.arange(0.0, 1.01, 0.1)

    assert cont_gen.line_type == line_type
    assert cont_gen.chunk_count == (1, 1)
    assert cont_gen.thread_count == 1

    for level in levels:
        lines = cont_gen.lines(level)
        util_test.assert_lines(lines, line_type)


@pytest.mark.parametrize("name", util_test.all_names())
@pytest.mark.parametrize("z", [np.nan, -np.nan, np.inf, -np.inf])
@pytest.mark.parametrize("zlevel", [0.0, np.nan, -np.nan, np.inf, -np.inf])
def test_lines_z_nonfinite(name: str, z: float, zlevel: float) -> None:
    cont_gen = contour_generator(z=[[z, z], [z, z]], name=name, line_type=LineType.SeparateCode)
    lines = cont_gen.lines(zlevel)
    assert lines == ([], [])



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.append(os.path.abspath("./sphinxext"))


# -- Project information -----------------------------------------------------

project = "ContourPy"
copyright = "2021-2023, ContourPy Contributors"
author = "ContourPy Contributors"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx_copybutton",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.extlinks",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "name_supports",
    "name_supports_type",
    "plot_directive",
]

autodoc_typehints = "none"

copybutton_prompt_text = r">>> |\.\.\. |\$ "
copybutton_prompt_is_regexp = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"

html_theme_path = ["_themes"]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_favicon = "_static/contourpy_favicon.ico"

html_theme_options = {
    "dark_logo": "contourpy_logo_horiz_white.svg",
    "light_logo": "contourpy_logo_horiz.svg",
    "sidebar_hide_name": True,
}

rst_epilog = """
.. _Bokeh: https://bokeh.org/
.. _Clang: https://clang.llvm.org/
.. _GCC: https://gcc.gnu.org/
.. _Matplotlib: https://matplotlib.org/
.. _MSVC: https://visualstudio.microsoft.com/
.. _NumPy: https://numpy.org/
.. _PyPI: https://pypi.org/project/contourpy/
.. _conda: https://conda.io/
.. _conda-forge: https://anaconda.org/conda-forge/contourpy/
.. _default: https://anaconda.org/anaconda/contourpy/
.. _github: https://www.github.com/contourpy/contourpy/
.. _meson: https://mesonbuild.com/
.. _meson-python: https://meson-python.readthedocs.io/
.. _ninja: https://ninja-build.org/
.. _pre-commit: https://pre-commit.com/
.. _pybind11: https://pybind11.readthedocs.io/
.. _pyenv: https://github.com/pyenv/pyenv/
.. _virtualenv: https://virtualenv.pypa.io/en/latest/
"""

extlinks = {
    "pr": ("https://github.com/contourpy/contourpy/pull/%s", "#%s"),
    "user": ("https://github.com/%s", "@%s"),
}



from __future__ import annotations

import re
from subprocess import run

from packaging.version import Version
import pytest

import contourpy


# From PEP440 appendix.
def version_is_canonical(version: str) -> bool:
    return re.match(
        r"^([1-9][0-9]*!)?(0|[1-9][0-9]*)(\.(0|[1-9][0-9]*))*((a|b|rc)(0|[1-9][0-9]*))?"
        r"(\.post(0|[1-9][0-9]*))?(\.dev(0|[1-9][0-9]*))?$", version) is not None


def test_cppcheck() -> None:
    # Skip test if cppcheck is not installed.
    cmd = ["cppcheck", "--version"]
    try:
        proc = run(cmd, capture_output=True)
    except FileNotFoundError:
        pytest.skip()

    cpp_version = Version(proc.stdout.decode("utf-8").strip().split()[-1])

    # Note excluding mpl2005 code.
    cmd = [
        "cppcheck", "--quiet", "--enable=all", "--error-exitcode=1", "src",
        "-isrc/mpl2005_original.cpp", "--suppress=missingIncludeSystem", "--inline-suppr"]

    if cpp_version >= Version("2.7"):
        cmd += ["--suppress=assertWithSideEffect"]

    proc = run(cmd, capture_output=True)
    assert proc.returncode == 0, f"cppcheck issues:\n{proc.stderr.decode('utf-8')}"


def test_mypy() -> None:
    # Skip test if mypy is not installed.
    cmd = ["mypy", "--version"]
    try:
        proc = run(cmd, capture_output=True)
    except FileNotFoundError:
        pytest.skip()

    cmd = ["mypy"]
    proc = run(cmd, capture_output=True)
    assert proc.returncode == 0, print(f"mypy issues:\n{proc.stdout.decode('utf-8')}")


def test_version() -> None:
    version_python = contourpy.__version__
    assert version_is_canonical(version_python)
    version_cxx = contourpy._contourpy.__version__
    assert version_python == version_cxx



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Sequence

import pytest

if TYPE_CHECKING:
    from _pytest.fixtures import SubRequest


def pytest_addoption(parser: pytest.Parser) -> None:
    parser.addoption(
        "--runslow", action="store_true", default=False, help="run slow tests",
    )
    parser.addoption(
        "--runtext", action="store_true", default=False, help="run tests with text output",
    )
    parser.addoption(
        "--driver-path", type=str, action="store", default="", help="path to chrome driver",
    )


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line("markers", "slow: mark test as slow to run")
    config.addinivalue_line("markers", "text: mark test as outputting text")
    config.addinivalue_line("markers", "image: mark test as generating comparison image")
    config.addinivalue_line("markers", "threads: mark test as using multiple threads")


def pytest_collection_modifyitems(config: pytest.Config, items: Sequence[Any]) -> None:
    if not config.getoption("--runslow"):
        skip_slow = pytest.mark.skip(reason="use --runslow option to run")
        for item in items:
            if "slow" in item.keywords:
                item.add_marker(skip_slow)

    if not config.getoption("--runtext"):
        skip_text = pytest.mark.skip(reason="use --runtext option to run")
        for item in items:
            if "text" in item.keywords and item.callspec.getparam("show_text"):
                item.add_marker(skip_text)


@pytest.fixture(scope="session")
def driver_path(request: SubRequest) -> Any:
    return request.config.getoption("--driver-path")



from __future__ import annotations

import sys
from typing import Any, ClassVar

from docutils.parsers.rst import Directive
from docutils.parsers.rst.directives import unchanged

import contourpy

sys.path.insert(0, ".")
from table import Table


class NameSupports(Directive):
    optional_arguments = 1

    option_spec: ClassVar[dict[str, Any]] = {  # type: ignore[misc]
        "filter": unchanged,
    }

    def run(self) -> list[Any]:
        names = list(contourpy._class_lookup)
        classes = list(contourpy._class_lookup.values())
        function_names = [
            "supports_corner_mask",
            "supports_quad_as_tri",
            "supports_threads",
            "supports_z_interp",
        ]

        filter_string = self.options.get("filter")
        if filter_string is not None:
            function_name = f"supports_{filter_string}"
            if function_name not in function_names:
                raise ValueError(f"Invalid filter string '{filter_string}'")
            function_names = [function_name]

        table = Table(1 + len(names))
        table.add_header(["", *names])

        for function_name in function_names:
            row = [function_name]
            for cls in classes:
                func = getattr(cls, function_name)
                if func():
                    row.append("Yes")
                else:
                    row.append("")
            table.add_row(row)

        return [table.get()]


def setup(app: Any) -> dict[str, bool]:
    app.add_directive("name_supports", NameSupports)
    return {"parallel_read_safe": True, "parallel_write_safe": True}



from __future__ import annotations

import pytest

from . import util_test


@pytest.mark.image
@pytest.mark.text
@pytest.mark.parametrize("show_text", [False, True])
@pytest.mark.parametrize("name", util_test.all_names())
def test_config_filled(show_text: bool, name: str) -> None:
    from . import util_config
    from .image_comparison import compare_images

    config = util_config.ConfigFilled(name, show_text=show_text)
    image_buffer = config.save_to_buffer()
    suffix = "" if show_text else "_no_text"
    compare_images(image_buffer, f"config_filled{suffix}.png", name)


@pytest.mark.image
@pytest.mark.text
@pytest.mark.parametrize("show_text", [False, True])
@pytest.mark.parametrize("name", util_test.quad_as_tri_names())
def test_config_filled_quad_as_tri(show_text: bool, name: str) -> None:
    from . import util_config
    from .image_comparison import compare_images

    config = util_config.ConfigFilled(name, quad_as_tri=True, show_text=show_text)
    image_buffer = config.save_to_buffer()
    suffix = "" if show_text else "_no_text"
    compare_images(image_buffer, f"config_filled_quad_as_tri{suffix}.png", name)


@pytest.mark.image
@pytest.mark.text
@pytest.mark.parametrize("show_text", [False, True])
@pytest.mark.parametrize("name", util_test.corner_mask_names())
def test_config_filled_corner(show_text: bool, name: str) -> None:
    from . import util_config
    from .image_comparison import compare_images

    config = util_config.ConfigFilledCorner(name, show_text=show_text)
    image_buffer = config.save_to_buffer()
    suffix = "" if show_text else "_no_text"
    compare_images(image_buffer, f"config_filled_corner{suffix}.png", name)


@pytest.mark.image
@pytest.mark.text
@pytest.mark.parametrize("show_text", [False, True])
@pytest.mark.parametrize("name", util_test.all_names())
def test_config_lines(show_text: bool, name: str) -> None:
    from . import util_config
    from .image_comparison import compare_images

    if name == "mpl2005":
        pytest.skip()  # Line directions are not consistent.
    config = util_config.ConfigLines(name, show_text=show_text)
    image_buffer = config.save_to_buffer()
    suffix = "" if show_text else "_no_text"
    compare_images(image_buffer, f"config_lines{suffix}.png", name)


@pytest.mark.image
@pytest.mark.text
@pytest.mark.parametrize("show_text", [False, True])
@pytest.mark.parametrize("name", util_test.quad_as_tri_names())
def test_config_lines_quad_as_tri(show_text: bool, name: str) -> None:
    from . import util_config
    from .image_comparison import compare_images

    config = util_config.ConfigLines(name, quad_as_tri=True, show_text=show_text)
    image_buffer = config.save_to_buffer()
    suffix = "" if show_text else "_no_text"
    compare_images(image_buffer, f"config_lines_quad_as_tri{suffix}.png", name)


@pytest.mark.image
@pytest.mark.text
@pytest.mark.parametrize("show_text", [False, True])
@pytest.mark.parametrize("name", util_test.corner_mask_names())
def test_config_lines_corner(show_text: bool, name: str) -> None:
    from . import util_config
    from .image_comparison import compare_images

    config = util_config.ConfigLinesCorner(name, show_text=show_text)
    image_buffer = config.save_to_buffer()
    suffix = "" if show_text else "_no_text"
    compare_images(image_buffer, f"config_lines_corner{suffix}.png", name)



from __future__ import annotations

from functools import reduce
from operator import add
from typing import TYPE_CHECKING, cast

import numpy as np
from numpy.testing import assert_allclose, assert_array_equal
import pytest

from contourpy import FillType, contour_generator
from contourpy.util.data import random, simple

from . import util_test

if TYPE_CHECKING:
    import contourpy._contourpy as cpy


@pytest.fixture
def two_outers_one_hole() -> tuple[cpy.CoordinateArray, ...]:
    x, y = np.meshgrid([0., 1., 2., 3.], [0., 1., 2.])
    z = np.array([[1.5, 1.5, 0.9, 0.0],
                  [1.5, 2.8, 0.4, 0.8],
                  [0.0, 0.0, 0.8, 1.9]])
    return x, y, z


@pytest.fixture
def xyz_chunk_test() -> tuple[cpy.CoordinateArray, ...]:
    x, y = np.meshgrid(np.arange(5), np.arange(5))
    z = 0.5*np.abs(y - 2) + 0.1*(x - 2)
    return x, y, z


@pytest.mark.parametrize("name", util_test.all_names())
def test_filled_decreasing_levels(name: str) -> None:
    cont_gen = contour_generator(z=[[0, 1], [2, 3]], name=name, fill_type=FillType.OuterCode)
    with pytest.raises(ValueError, match="upper and lower levels are the wrong way round"):
        cont_gen.filled(2.0, 1.0)


@pytest.mark.image
@pytest.mark.parametrize("name, fill_type", util_test.all_names_and_fill_types())
def test_filled_simple(name: str, fill_type: FillType) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = simple((30, 40))
    cont_gen = contour_generator(x, y, z, name=name, fill_type=fill_type)
    levels = np.arange(-1.0, 1.01, 0.1)

    assert cont_gen.fill_type == fill_type
    assert cont_gen.chunk_count == (1, 1)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)-1):
        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "filled_simple.png", f"{name}_{fill_type}")


@pytest.mark.image
@pytest.mark.parametrize("name, fill_type", util_test.all_names_and_fill_types())
def test_filled_simple_chunk(name: str, fill_type: FillType) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = simple((30, 40))
    cont_gen = contour_generator(
        x, y, z, name=name, fill_type=fill_type, chunk_size=2, thread_count=1,
    )
    levels = np.arange(-1.0, 1.01, 0.1)

    assert cont_gen.fill_type == fill_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)-1):
        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(
        image_buffer, "filled_simple_chunk.png", f"{name}_{fill_type}", mean_threshold=0.12,
    )


@pytest.mark.image
@pytest.mark.threads
@pytest.mark.parametrize("fill_type", FillType.__members__.values())
@pytest.mark.parametrize("thread_count", util_test.thread_counts())
def test_filled_simple_chunk_threads(fill_type: FillType, thread_count: int) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = simple((30, 40))
    cont_gen = contour_generator(
        x, y, z, name="threaded", fill_type=fill_type, chunk_size=2, thread_count=thread_count,
    )
    levels = np.arange(-1.0, 1.01, 0.1)

    assert cont_gen.fill_type == fill_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == thread_count

    renderer = MplTestRenderer()
    for i in range(len(levels)-1):
        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(
        image_buffer, "filled_simple_chunk.png", f"{fill_type}_{thread_count}", mean_threshold=0.12,
    )


@pytest.mark.image
@pytest.mark.parametrize("name, fill_type", util_test.all_names_and_fill_types())
def test_filled_simple_no_corner_mask(name: str, fill_type: FillType) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = simple((30, 40), want_mask=True)
    cont_gen = contour_generator(x, y, z, name=name, fill_type=fill_type, corner_mask=False)
    levels = np.arange(-1.0, 1.01, 0.1)

    assert cont_gen.fill_type == fill_type
    assert cont_gen.chunk_count == (1, 1)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)-1):
        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "filled_simple_no_corner_mask.png", f"{name}_{fill_type}")


@pytest.mark.image
@pytest.mark.parametrize("name, fill_type", util_test.all_names_and_fill_types())
def test_filled_simple_no_corner_mask_chunk(name: str, fill_type: FillType) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = simple((30, 40), want_mask=True)
    cont_gen = contour_generator(
        x, y, z, name=name, fill_type=fill_type, corner_mask=False, chunk_size=2, thread_count=1,
    )
    levels = np.arange(-1.0, 1.01, 0.1)

    assert cont_gen.fill_type == fill_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)-1):
        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(
        image_buffer, "filled_simple_no_corner_mask_chunk.png", f"{name}_{fill_type}",
        mean_threshold=0.11,
    )


@pytest.mark.image
@pytest.mark.threads
@pytest.mark.parametrize("fill_type", FillType.__members__.values())
@pytest.mark.parametrize("thread_count", util_test.thread_counts())
def test_filled_simple_no_corner_mask_chunk_threads(fill_type: FillType, thread_count: int) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = simple((30, 40), want_mask=True)
    cont_gen = contour_generator(
        x, y, z, name="threaded", fill_type=fill_type, corner_mask=False, chunk_size=2,
        thread_count=thread_count,
    )
    levels = np.arange(-1.0, 1.01, 0.1)

    assert cont_gen.fill_type == fill_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == thread_count

    renderer = MplTestRenderer()
    for i in range(len(levels)-1):
        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(
        image_buffer, "filled_simple_no_corner_mask_chunk.png", f"{fill_type}_{thread_count}",
        mean_threshold=0.11,
    )


@pytest.mark.image
@pytest.mark.parametrize("name", util_test.corner_mask_names())
def test_filled_simple_corner_mask(name: str) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = simple((30, 40), want_mask=True)
    fill_type = FillType.OuterCode
    cont_gen = contour_generator(x, y, z, name=name, fill_type=fill_type, corner_mask=True)
    levels = np.arange(-1.0, 1.01, 0.1)

    assert cont_gen.fill_type == fill_type
    assert cont_gen.chunk_count == (1, 1)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)-1):
        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "filled_simple_corner_mask.png", f"{name}_{fill_type}")


@pytest.mark.image
@pytest.mark.parametrize("name", util_test.corner_mask_names())
def test_filled_simple_corner_mask_chunk(name: str) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = simple((30, 40), want_mask=True)
    fill_type = FillType.OuterCode
    cont_gen = contour_generator(
        x, y, z, name=name, fill_type=fill_type, corner_mask=True, chunk_size=2, thread_count=1,
    )
    levels = np.arange(-1.0, 1.01, 0.1)

    assert cont_gen.fill_type == fill_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)-1):
        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(
        image_buffer, "filled_simple_corner_mask_chunk.png", f"{name}_{fill_type}",
        mean_threshold=0.10,
    )


@pytest.mark.image
@pytest.mark.threads
@pytest.mark.parametrize("fill_type", FillType.__members__.values())
@pytest.mark.parametrize("thread_count", util_test.thread_counts())
def test_filled_simple_corner_mask_chunk_threads(fill_type: FillType, thread_count: int) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = simple((30, 40), want_mask=True)
    fill_type = FillType.OuterCode
    cont_gen = contour_generator(
        x, y, z, name="threaded", fill_type=fill_type, corner_mask=True, chunk_size=2,
        thread_count=thread_count,
    )
    levels = np.arange(-1.0, 1.01, 0.1)

    assert cont_gen.fill_type == fill_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == thread_count

    renderer = MplTestRenderer()
    for i in range(len(levels)-1):
        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(
        image_buffer, "filled_simple_corner_mask_chunk.png", f"{fill_type}_{thread_count}",
        mean_threshold=0.10,
    )


@pytest.mark.image
@pytest.mark.parametrize("name", util_test.quad_as_tri_names())
def test_filled_simple_quad_as_tri(name: str) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = simple((30, 40))
    cont_gen = contour_generator(x, y, z, name=name, quad_as_tri=True)
    levels = np.arange(-1.0, 1.01, 0.1)

    assert cont_gen.chunk_count == (1, 1)
    assert cont_gen.thread_count == 1
    assert cont_gen.quad_as_tri

    fill_type = cont_gen.fill_type
    renderer = MplTestRenderer()
    for i in range(len(levels)-1):
        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "filled_simple_quad_as_tri.png", f"{name}")


@pytest.mark.image
@pytest.mark.parametrize("name, fill_type", util_test.all_names_and_fill_types())
def test_filled_random(name: str, fill_type: FillType) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = random((30, 40), mask_fraction=0.0)
    cont_gen = contour_generator(x, y, z, name=name, fill_type=fill_type)
    levels = np.arange(0.0, 1.01, 0.2)

    assert cont_gen.fill_type == fill_type
    assert cont_gen.chunk_count == (1, 1)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)-1):
        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "filled_random.png", f"{name}_{fill_type}")


@pytest.mark.image
@pytest.mark.parametrize("name, fill_type", util_test.all_names_and_fill_types())
def test_filled_random_chunk(name: str, fill_type: FillType) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = random((30, 40), mask_fraction=0.0)
    cont_gen = contour_generator(
        x, y, z, name=name, fill_type=fill_type, chunk_size=2, thread_count=1,
    )
    levels = np.arange(0.0, 1.01, 0.2)

    assert cont_gen.fill_type == fill_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)-1):
        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    max_threshold = None
    mean_threshold = None
    if name == "mpl2005":
        max_threshold = 128
        mean_threshold = 0.16
    elif name in ("serial", "threaded"):
        if fill_type in (FillType.ChunkCombinedCode, FillType.ChunkCombinedOffset):
            max_threshold = 99
            mean_threshold = 0.142
        else:
            max_threshold = 135
            mean_threshold = 0.19

    compare_images(
        image_buffer, "filled_random_chunk.png", f"{name}_{fill_type}",
        max_threshold=max_threshold, mean_threshold=mean_threshold,
    )


@pytest.mark.image
@pytest.mark.threads
@pytest.mark.parametrize("fill_type", FillType.__members__.values())
@pytest.mark.parametrize("thread_count", util_test.thread_counts())
def test_filled_random_chunk_threads(fill_type: FillType, thread_count: int) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = random((30, 40), mask_fraction=0.0)
    cont_gen = contour_generator(
        x, y, z, name="threaded", fill_type=fill_type, chunk_size=2, thread_count=thread_count,
    )
    levels = np.arange(0.0, 1.01, 0.2)

    assert cont_gen.fill_type == fill_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == thread_count

    renderer = MplTestRenderer()
    for i in range(len(levels)-1):
        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    if fill_type in (FillType.ChunkCombinedCode, FillType.ChunkCombinedOffset):
        max_threshold = 99
        mean_threshold = 0.142
    else:
        max_threshold = 135
        mean_threshold = 0.19

    compare_images(
        image_buffer, "filled_random_chunk.png", f"{fill_type}_{thread_count}",
        max_threshold=max_threshold, mean_threshold=mean_threshold,
    )


@pytest.mark.image
@pytest.mark.parametrize("name, fill_type", util_test.all_names_and_fill_types())
def test_filled_random_no_corner_mask(name: str, fill_type: FillType) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = random((30, 40), mask_fraction=0.05)
    cont_gen = contour_generator(x, y, z, name=name, fill_type=fill_type, corner_mask=False)
    levels = np.arange(0.0, 1.01, 0.2)

    assert cont_gen.fill_type == fill_type
    assert cont_gen.chunk_count == (1, 1)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)-1):
        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "filled_random_no_corner_mask.png", f"{name}_{fill_type}")


@pytest.mark.image
@pytest.mark.parametrize("name, fill_type", util_test.all_names_and_fill_types())
def test_filled_random_no_corner_mask_chunk(name: str, fill_type: FillType) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = random((30, 40), mask_fraction=0.05)
    cont_gen = contour_generator(
        x, y, z, name=name, fill_type=fill_type, corner_mask=False, chunk_size=2, thread_count=1,
    )
    levels = np.arange(0.0, 1.01, 0.2)

    assert cont_gen.fill_type == fill_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)-1):
        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    max_threshold = None
    mean_threshold = None
    if name == "mpl2005":
        max_threshold = 128
        mean_threshold = 0.19
    elif name in ("serial", "threaded"):
        if fill_type in (FillType.ChunkCombinedCode, FillType.ChunkCombinedOffset):
            max_threshold = 99
            mean_threshold = 0.18
        else:
            max_threshold = 135
            mean_threshold = 0.23

    compare_images(
        image_buffer, "filled_random_no_corner_mask_chunk.png", f"{name}_{fill_type}",
        max_threshold=max_threshold, mean_threshold=mean_threshold,
    )


@pytest.mark.image
@pytest.mark.threads
@pytest.mark.parametrize("fill_type", FillType.__members__.values())
@pytest.mark.parametrize("thread_count", util_test.thread_counts())
def test_filled_random_no_corner_mask_chunk_threads(fill_type: FillType, thread_count: int) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = random((30, 40), mask_fraction=0.05)
    cont_gen = contour_generator(
        x, y, z, name="threaded", fill_type=fill_type, corner_mask=False, chunk_size=2,
        thread_count=thread_count,
    )
    levels = np.arange(0.0, 1.01, 0.2)

    assert cont_gen.fill_type == fill_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == thread_count

    renderer = MplTestRenderer()
    for i in range(len(levels)-1):
        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    if fill_type in (FillType.ChunkCombinedCode, FillType.ChunkCombinedOffset):
        max_threshold = 99
        mean_threshold = 0.18
    else:
        max_threshold = 135
        mean_threshold = 0.23

    compare_images(
        image_buffer, "filled_random_no_corner_mask_chunk.png", f"{fill_type}_{thread_count}",
        max_threshold=max_threshold, mean_threshold=mean_threshold,
    )


@pytest.mark.image
@pytest.mark.parametrize("name", util_test.corner_mask_names())
def test_filled_random_corner_mask(name: str) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = random((30, 40), mask_fraction=0.05)
    fill_type = FillType.OuterCode
    cont_gen = contour_generator(x, y, z, name=name, corner_mask=True, fill_type=fill_type)
    levels = np.arange(0.0, 1.01, 0.2)

    assert cont_gen.fill_type == fill_type
    assert cont_gen.chunk_count == (1, 1)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)-1):
        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "filled_random_corner_mask.png", f"{name}_{fill_type}")


@pytest.mark.image
@pytest.mark.parametrize("name", util_test.corner_mask_names())
def test_filled_random_corner_mask_chunk(name: str) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = random((30, 40), mask_fraction=0.05)
    fill_type = FillType.OuterCode
    cont_gen = contour_generator(
        x, y, z, name=name, corner_mask=True, fill_type=fill_type, chunk_size=2, thread_count=1,
    )
    levels = np.arange(0.0, 1.01, 0.2)

    assert cont_gen.fill_type == fill_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == 1

    renderer = MplTestRenderer()
    for i in range(len(levels)-1):
        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    max_threshold = None
    mean_threshold = None
    if name in ("serial", "threaded"):
        max_threshold = 135
        mean_threshold = 0.17

    compare_images(
        image_buffer, "filled_random_corner_mask_chunk.png", f"{name}_{fill_type}",
        max_threshold=max_threshold, mean_threshold=mean_threshold,
    )


@pytest.mark.image
@pytest.mark.threads
@pytest.mark.parametrize("fill_type", FillType.__members__.values())
@pytest.mark.parametrize("thread_count", util_test.thread_counts())
def test_filled_random_corner_mask_chunk_threads(fill_type: FillType, thread_count: int) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = random((30, 40), mask_fraction=0.05)
    fill_type = FillType.OuterCode
    cont_gen = contour_generator(
        x, y, z, name="threaded", corner_mask=True, fill_type=fill_type, chunk_size=2,
        thread_count=thread_count,
    )
    levels = np.arange(0.0, 1.01, 0.2)

    assert cont_gen.fill_type == fill_type
    assert cont_gen.chunk_count == (15, 20)
    assert cont_gen.thread_count == thread_count

    renderer = MplTestRenderer()
    for i in range(len(levels)-1):
        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(
        image_buffer, "filled_random_corner_mask_chunk.png", f"{fill_type}_{thread_count}",
        max_threshold=135, mean_threshold=0.17,
    )


@pytest.mark.image
@pytest.mark.parametrize("name", util_test.quad_as_tri_names())
def test_filled_random_quad_as_tri(name: str) -> None:
    from contourpy.util.mpl_renderer import MplTestRenderer

    from .image_comparison import compare_images

    x, y, z = random((30, 40), mask_fraction=0.0)
    cont_gen = contour_generator(x, y, z, name=name, quad_as_tri=True)
    levels = np.arange(0.0, 1.01, 0.2)

    assert cont_gen.thread_count == 1
    assert cont_gen.quad_as_tri

    fill_type = cont_gen.fill_type
    renderer = MplTestRenderer()
    for i in range(len(levels)-1):
        renderer.filled(cont_gen.filled(levels[i], levels[i+1]), fill_type, color=f"C{i}")
    image_buffer = renderer.save_to_buffer()

    compare_images(image_buffer, "filled_random_quad_as_tri.png", f"{name}")


@pytest.mark.parametrize("fill_type", FillType.__members__.values())
@pytest.mark.parametrize("name", ["serial", "threaded"])
def test_return_by_fill_type(
    two_outers_one_hole: tuple[cpy.CoordinateArray, ...],
    name: str,
    fill_type: FillType,
) -> None:
    x, y, z = two_outers_one_hole
    cont_gen = contour_generator(x, y, z, name=name, fill_type=fill_type)

    assert cont_gen.fill_type == fill_type
    assert cont_gen.thread_count == 1

    filled = cont_gen.filled(1.0, 2.0)

    util_test.assert_filled(filled, fill_type)

    # Helper functions to avoid code duplication for the different FillTypes.
    def assert_outer_points(points: list[cpy.PointArray]) -> None:
        assert isinstance(points, list) and len(points) == 2
        assert points[0].shape == (13, 2)
        assert points[1].shape == (4, 2)
        assert_array_equal(points[0][0], points[0][7])
        assert_array_equal(points[0][8], points[0][12])
        assert_array_equal(points[1][0], points[1][3])

    def assert_chunk_points(points_or_none: list[cpy.PointArray | None]) -> None:
        assert isinstance(points_or_none, list) and len(points_or_none) == 1
        assert points_or_none[0] is not None
        points = points_or_none[0]
        assert points.shape == (17, 2)
        assert_array_equal(points[0], points[7])
        assert_array_equal(points[8], points[12])
        assert_array_equal(points[13], points[16])

    def assert_chunk_codes(codes_or_none: list[cpy.CodeArray | None]) -> None:
        assert isinstance(codes_or_none, list) and len(codes_or_none) == 1
        assert codes_or_none[0] is not None
        codes = codes_or_none[0]
        assert_array_equal(codes, [1, 2, 2, 2, 2, 2, 2, 79, 1, 2, 2, 2, 79, 1, 2, 2, 79])

    def assert_chunk_offsets(offsets_or_none: list[cpy.OffsetArray | None]) -> None:
        assert isinstance(offsets_or_none, list) and len(offsets_or_none) == 1
        assert offsets_or_none[0] is not None
        offsets = offsets_or_none[0]
        assert_array_equal(offsets, [0, 8, 13, 17])

    if fill_type == FillType.OuterCode:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_OuterCode, filled)
        assert_outer_points(filled[0])
        codes = filled[1]
        assert isinstance(codes, list) and len(codes) == 2
        assert_array_equal(codes[0], [1, 2, 2, 2, 2, 2, 2, 79, 1, 2, 2, 2, 79])
        assert_array_equal(codes[1], [1, 2, 2, 79])
    elif fill_type == FillType.OuterOffset:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_OuterOffset, filled)
        assert_outer_points(filled[0])
        offsets = filled[1]
        assert isinstance(offsets, list) and len(offsets) == 2
        assert_array_equal(offsets[0], [0, 8, 13])
        assert_array_equal(offsets[1], [0, 4])
    elif fill_type == FillType.ChunkCombinedCode:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_ChunkCombinedCode, filled)
        assert_chunk_points(filled[0])
        assert_chunk_codes(filled[1])
    elif fill_type == FillType.ChunkCombinedOffset:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_ChunkCombinedOffset, filled)
        assert_chunk_points(filled[0])
        assert_chunk_offsets(filled[1])
    elif fill_type == FillType.ChunkCombinedCodeOffset:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_ChunkCombinedCodeOffset, filled)
        assert_chunk_points(filled[0])
        assert_chunk_codes(filled[1])

        outer_offsets_or_none = filled[2]
        assert isinstance(outer_offsets_or_none, list) and len(outer_offsets_or_none) == 1
        assert outer_offsets_or_none[0] is not None
        assert_array_equal(outer_offsets_or_none[0], [0, 13, 17])
    elif fill_type == FillType.ChunkCombinedOffsetOffset:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_ChunkCombinedOffsetOffset, filled)
        assert_chunk_points(filled[0])
        assert_chunk_offsets(filled[1])

        outer_offsets_or_none = filled[2]
        assert isinstance(outer_offsets_or_none, list) and len(outer_offsets_or_none) == 1
        assert outer_offsets_or_none[0] is not None
        assert_array_equal(outer_offsets_or_none[0], [0, 2, 3])
    else:
        raise RuntimeError(f"Unexpected fill_type {fill_type}")


@pytest.mark.threads
@pytest.mark.parametrize("fill_type", FillType.__members__.values())
@pytest.mark.parametrize("name, thread_count",
                         [("serial", 1), ("threaded", 1), ("threaded", 2)])
def test_return_by_fill_type_chunk(
    xyz_chunk_test: tuple[cpy.CoordinateArray, ...],
    name: str,
    thread_count: int,
    fill_type: FillType,
) -> None:
    x, y, z = xyz_chunk_test
    cont_gen = contour_generator(
        x, y, z, name=name, fill_type=fill_type, chunk_count=2, thread_count=thread_count,
    )

    assert cont_gen.fill_type == fill_type
    assert cont_gen.chunk_count == (2, 2)
    assert cont_gen.chunk_size == (2, 2)
    assert cont_gen.thread_count == thread_count

    filled = cont_gen.filled(0.45, 0.55)

    util_test.assert_filled(filled, fill_type)

    # Expected points by chunk.
    expected: list[cpy.PointArray] = [
        np.asarray([[0.0, 0.7], [0.0, 0.5], [1.0, 0.7], [2.0, 0.9], [2.0, 1.0], [2.0, 1.1],
                    [1.5, 1.0], [1.0, 0.9], [0.0, 0.7]]),
        np.asarray([[2.0, 1.0], [2.0, 0.9], [2.5, 1.0], [3.0, 1.1], [4.0, 1.3], [4.0, 1.5],
                    [3.0, 1.3], [2.0, 1.1], [2.0, 1.0]]),
        np.asarray([[1.5, 3.0], [2.0, 2.9], [2.0, 3.0], [2.0, 3.1], [1.0, 3.3], [0.0, 3.5],
                    [0.0, 3.3], [1.0, 3.1], [1.5, 3.0]]),
        np.asarray([[2.0, 3.0], [2.0, 2.9], [3.0, 2.7], [4.0, 2.5], [4.0, 2.7], [3.0, 2.9],
                    [2.5, 3.0], [2.0, 3.1], [2.0, 3.0]]),
    ]

    # Helper functions to avoid code duplication for the different FillTypes.
    def assert_outer_points(points: list[cpy.PointArray], expected: list[cpy.PointArray]) -> None:
        assert isinstance(points, list) and len(points) == 4
        if name == "threaded" and cont_gen.thread_count > 1:
            # Polygons may be in any order so sort lines and expected.
            points = util_test.sort_by_first_xy(points)
            expected = util_test.sort_by_first_xy(expected)
        for chunk in range(4):
            assert_allclose(points[chunk], expected[chunk])

    def assert_chunk_points(
        points_or_none: list[cpy.PointArray | None], expected: list[cpy.PointArray],
    ) -> None:
        assert isinstance(points_or_none, list) and len(points_or_none) == 4
        for chunk in range(4):
            chunk_points = points_or_none[chunk]
            assert chunk_points is not None
            assert_allclose(chunk_points, expected[chunk])

    def assert_chunk_codes(codes_or_none: list[cpy.CodeArray | None]) -> None:
        assert isinstance(codes_or_none, list) and len(codes_or_none) == 4
        for chunk in range(4):
            chunk_codes = codes_or_none[chunk]
            assert chunk_codes is not None
            assert_array_equal(chunk_codes, [1, 2, 2, 2, 2, 2, 2, 2, 79])

    def assert_chunk_offsets(offsets_or_none: list[cpy.OffsetArray | None]) -> None:
        assert isinstance(offsets_or_none, list) and len(offsets_or_none) == 4
        for chunk in range(4):
            chunk_offsets = offsets_or_none[chunk]
            assert chunk_offsets is not None
            assert_array_equal(chunk_offsets, [0, 9])

    if fill_type == FillType.OuterCode:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_OuterCode, filled)
        assert_outer_points(filled[0], expected)
        codes = filled[1]
        assert isinstance(codes, list) and len(codes) == 4
        for chunk in range(4):
            assert_array_equal(codes[chunk], [1, 2, 2, 2, 2, 2, 2, 2, 79])
    elif fill_type == FillType.OuterOffset:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_OuterOffset, filled)
        assert_outer_points(filled[0], expected)
        offsets = filled[1]
        assert isinstance(offsets, list) and len(offsets) == 4
        for chunk in range(4):
            assert_array_equal(offsets[chunk], [0, 9])
    elif fill_type == FillType.ChunkCombinedCode:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_ChunkCombinedCode, filled)
        assert_chunk_points(filled[0], expected)
        assert_chunk_codes(filled[1])
    elif fill_type == FillType.ChunkCombinedOffset:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_ChunkCombinedOffset, filled)
        assert_chunk_points(filled[0], expected)
        assert_chunk_offsets(filled[1])
    elif fill_type == FillType.ChunkCombinedCodeOffset:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_ChunkCombinedCodeOffset, filled)
        assert_chunk_points(filled[0], expected)
        assert_chunk_codes(filled[1])

        outer_offsets_or_none = filled[2]
        assert isinstance(outer_offsets_or_none, list) and len(outer_offsets_or_none) == 4
        for chunk in range(4):
            chunk_outer_offsets = outer_offsets_or_none[chunk]
            assert chunk_outer_offsets is not None
            assert_array_equal(chunk_outer_offsets, [0, 9])
    elif fill_type == FillType.ChunkCombinedOffsetOffset:
        if TYPE_CHECKING:
            filled = cast(cpy.FillReturn_ChunkCombinedOffsetOffset, filled)
        assert_chunk_points(filled[0], expected)
        assert_chunk_offsets(filled[1])

        outer_offsets_or_none = filled[2]
        assert isinstance(outer_offsets_or_none, list) and len(outer_offsets_or_none) == 4
        for chunk in range(4):
            chunk_outer_offsets = outer_offsets_or_none[chunk]
            assert chunk_outer_offsets is not None
            assert_array_equal(chunk_outer_offsets, [0, 1])
    else:
        raise RuntimeError(f"Unexpected fill_type {fill_type}")


@pytest.mark.parametrize("name, fill_type", util_test.all_names_and_fill_types())
@pytest.mark.parametrize("corner_mask", [None, False, True])
def test_filled_random_big(name: str, fill_type: FillType, corner_mask: bool | None) -> None:
    if corner_mask and name in ["mpl2005", "mpl2014"]:
        pytest.skip()

    x, y, z = random((1000, 1000), mask_fraction=0.0 if corner_mask is None else 0.05)
    cont_gen = contour_generator(
        x, y, z, name=name, corner_mask=corner_mask if isinstance(corner_mask, bool) else False,
        fill_type=fill_type,
    )
    levels = np.arange(0.0, 1.01, 0.1)

    assert cont_gen.fill_type == fill_type

    for i in range(len(levels)-1):
        filled = cont_gen.filled(levels[i], levels[i+1])
        util_test.assert_filled(filled, fill_type)


@pytest.mark.slow
@pytest.mark.parametrize("seed", np.arange(10))
def test_filled_compare_slow(seed: int) -> None:
    x, y, z = random((1000, 1000), mask_fraction=0.05, seed=seed)
    levels = np.arange(0.0, 1.01, 0.1)

    cont_gen_mpl2014 = contour_generator(x, y, z, name="mpl2014", corner_mask=True)
    cont_gen_serial = contour_generator(
        x, y, z, name="serial", corner_mask=True, fill_type=FillType.ChunkCombinedOffsetOffset)

    assert cont_gen_mpl2014.fill_type == FillType.OuterCode
    assert cont_gen_serial.fill_type == FillType.ChunkCombinedOffsetOffset

    for i in range(len(levels)-1):
        filled_mpl2014 = cont_gen_mpl2014.filled(levels[i], levels[i+1])
        util_test.assert_filled(filled_mpl2014, cont_gen_mpl2014.fill_type)

        filled_serial = cont_gen_serial.filled(levels[i], levels[i+1])
        util_test.assert_filled(filled_serial, cont_gen_serial.fill_type)

        if TYPE_CHECKING:
            filled_mpl2014 = cast(cpy.FillReturn_OuterCode, filled_mpl2014)
            filled_serial = cast(cpy.FillReturn_ChunkCombinedOffsetOffset, filled_serial)

        # Check same results obtained for each in terms of number of points, etc.
        code_list = filled_mpl2014[1]
        n_points = reduce(add, map(len, code_list))
        n_outers = len(code_list)
        n_lines = reduce(add, map(lambda c: np.count_nonzero(c == 1), code_list))

        assert filled_serial[0][0] is not None
        assert n_points == len(filled_serial[0][0])

        assert filled_serial[1][0] is not None
        assert n_lines == len(filled_serial[1][0]) - 1

        assert filled_serial[2][0] is not None
        assert n_outers == len(filled_serial[2][0]) - 1


@pytest.mark.parametrize("name", util_test.all_names())
@pytest.mark.parametrize("z", [np.nan, -np.nan, np.inf, -np.inf])
@pytest.mark.parametrize("zlevel", [0.0, np.nan, -np.nan, np.inf, -np.inf])
def test_filled_z_nonfinite(name: str, z: float, zlevel: float) -> None:
    cont_gen = contour_generator(z=[[z, z], [z, z]], name=name, fill_type=FillType.OuterCode)
    filled = cont_gen.filled(zlevel, zlevel)
    assert filled == ([], [])



# Test internal API defined in wrap.cpp, mostly constructors. These are not normally called from
# client code as contour_generator() defined in Python is preferred.

from __future__ import annotations

from io import StringIO
import sys
from typing import TYPE_CHECKING

import pytest

from contourpy._contourpy import (
    ContourGenerator, FillType, LineType, Mpl2005ContourGenerator, Mpl2014ContourGenerator,
    SerialContourGenerator, ThreadedContourGenerator, ZInterp,
)

if TYPE_CHECKING:
    from typing_extensions import TypeAlias

    KWArgs: TypeAlias = dict[str, int | bool | LineType | FillType | ZInterp]
    XYZMask: TypeAlias = tuple[list[list[int]], list[list[int]], list[list[int]], None]


def all_classes() -> list[type[ContourGenerator]]:
    return [Mpl2005ContourGenerator, Mpl2014ContourGenerator, SerialContourGenerator,
            ThreadedContourGenerator]


def default_kwargs(cls: type[ContourGenerator]) -> KWArgs:
    kwargs: KWArgs = {
        "x_chunk_size": 0,
        "y_chunk_size": 0,
    }

    if cls == Mpl2005ContourGenerator:
        pass
    elif cls == Mpl2014ContourGenerator:
        kwargs["corner_mask"] = True
    else:
        kwargs.update({
            "line_type": LineType.Separate,
            "fill_type": FillType.OuterOffset,
            "corner_mask": True,
            "quad_as_tri": False,
            "z_interp": ZInterp.Linear,
        })
        if cls == ThreadedContourGenerator:
            kwargs["thread_count"] = 1
    return kwargs


@pytest.fixture
def xyz_mask() -> XYZMask:
    x = [[0, 1], [0, 1]]
    y = [[0, 0], [1, 1]]
    z = [[1, 2], [3, 4]]
    return x, y, z, None


@pytest.mark.parametrize("cls", all_classes())
def test_default(cls: type[ContourGenerator], xyz_mask: XYZMask) -> None:
    kwargs = default_kwargs(cls)
    cls(*xyz_mask, **kwargs)


@pytest.mark.parametrize("cls", all_classes())
def test_xyz_ndim(cls: type[ContourGenerator], xyz_mask: XYZMask) -> None:
    x, y, z, mask = xyz_mask
    one_d = [1]
    three_d = [[[1]]]
    kwargs = default_kwargs(cls)
    msg = "x, y and z must all be 2D arrays"

    with pytest.raises(ValueError, match=msg):
        cls(one_d, y, z, mask, **kwargs)
    with pytest.raises(ValueError, match=msg):
        cls(three_d, y, z, mask, **kwargs)

    with pytest.raises(ValueError, match=msg):
        cls(x, one_d, z, mask, **kwargs)
    with pytest.raises(ValueError, match=msg):
        cls(x, three_d, z, mask, **kwargs)

    with pytest.raises(ValueError, match=msg):
        cls(x, y, one_d, mask, **kwargs)
    with pytest.raises(ValueError, match=msg):
        cls(x, y, three_d, mask, **kwargs)


@pytest.mark.parametrize("cls", all_classes())
def test_xyz_shape(cls: type[ContourGenerator], xyz_mask: XYZMask) -> None:
    x, y, z, mask = xyz_mask
    diff_shape = [[0, 1], [2, 3], [4, 5]]
    kwargs = default_kwargs(cls)
    msg = "x, y and z arrays must have the same shape"

    with pytest.raises(ValueError, match=msg):
        cls(diff_shape, y, z, mask, **kwargs)
    with pytest.raises(ValueError, match=msg):
        cls(x, diff_shape, z, mask, **kwargs)
    with pytest.raises(ValueError, match=msg):
        cls(x, y, diff_shape, mask, **kwargs)


@pytest.mark.parametrize("cls", all_classes())
def test_xy_at_least_2x2(cls: type[ContourGenerator]) -> None:
    kwargs = default_kwargs(cls)
    msg = "x, y and z must all be at least 2x2 arrays"

    with pytest.raises(ValueError, match=msg):
        cls([[1, 2]], [[1, 2]], [[1, 2]], None, **kwargs)
    with pytest.raises(ValueError, match=msg):
        cls([[1], [2]], [[1], [2]], [[1], [2]], None, **kwargs)


@pytest.mark.parametrize("cls", all_classes())
def test_mask_2d(cls: type[ContourGenerator], xyz_mask: XYZMask) -> None:
    x, y, z, _ = xyz_mask
    kwargs = default_kwargs(cls)
    msg = "mask array must be a 2D array"

    with pytest.raises(ValueError, match=msg):
        cls(x, y, z, [False, True], **kwargs)
    with pytest.raises(ValueError, match=msg):
        cls(x, y, z, [[[False, True]]], **kwargs)


@pytest.mark.parametrize("cls", all_classes())
def test_mask_shape(cls: type[ContourGenerator], xyz_mask: XYZMask) -> None:
    x, y, z, _ = xyz_mask
    kwargs = default_kwargs(cls)
    msg = "If mask is set it must be a 2D array with the same shape as z"

    with pytest.raises(ValueError, match=msg):
        cls(x, y, z, [[False, True]], **kwargs)
    with pytest.raises(ValueError, match=msg):
        cls(x, y, z, [[False], [True]], **kwargs)


@pytest.mark.parametrize("cls", all_classes())
def test_chunk_size_not_negative(cls: type[ContourGenerator], xyz_mask: XYZMask) -> None:
    x, y, z, mask = xyz_mask
    kwargs = default_kwargs(cls)
    msg = "x_chunk_size and y_chunk_size cannot be negative"

    x_kwargs = kwargs.copy()
    x_kwargs.update({"x_chunk_size": -1})
    with pytest.raises(ValueError, match=msg):
        cls(x, y, z, mask, **x_kwargs)

    y_kwargs = kwargs.copy()
    y_kwargs.update({"y_chunk_size": -1})
    with pytest.raises(ValueError, match=msg):
        cls(x, y, z, mask, **y_kwargs)


@pytest.mark.parametrize("cls", [SerialContourGenerator, ThreadedContourGenerator])
def test_log_z_not_negative(cls: type[ContourGenerator], xyz_mask: XYZMask) -> None:
    x, y, _, mask = xyz_mask
    kwargs = default_kwargs(cls)
    kwargs["z_interp"] = ZInterp.Log
    msg = "z values must be positive if using ZInterp.Log"

    z = [[0, 1],[2, 3]]
    with pytest.raises(ValueError, match=msg):
        cls(x, y, z, mask, **kwargs)

    # Ignore z that are masked out
    mask = [[True, False], [False, False]]
    cls(x, y, z, mask, **kwargs)


@pytest.mark.skipif(sys.platform not in ("darwin", "linux"), reason="sysout redirect not available")
@pytest.mark.parametrize("cls", [SerialContourGenerator, ThreadedContourGenerator])
def test_write_cache(cls: type[ContourGenerator], xyz_mask: XYZMask) -> None:
    from wurlitzer import STDOUT, pipes

    x, y, z, _ = xyz_mask
    mask = [[False, False], [False, True]]
    kwargs = default_kwargs(cls)

    cg = cls(x, y, z, mask, **kwargs)

    cg.filled(2.5, 3.5)
    out = StringIO()
    with pipes(stdout=out, stderr=STDOUT):
        cg._write_cache()  # type: ignore[attr-defined]
    assert out.getvalue() == \
        "---------- Cache ----------\n" \
        "j=1 ...e10....... .SW.20.w..... \n" \
        "j=0 ....00....... ...n00....... \n" \
        "    i=0           i=1           \n" \
        "---------------------------\n"

    cg.lines(2.5)
    out = StringIO()
    with pipes(stdout=out, stderr=STDOUT):
        cg._write_cache()  # type: ignore[attr-defined]
    assert out.getvalue() == \
        "---------- Cache ----------\n" \
        "j=1 ...e10....... .SW.10.w..... \n" \
        "j=0 ....00....... ...n00....... \n" \
        "    i=0           i=1           \n" \
        "---------------------------\n"



from __future__ import annotations

import os
from typing import Any, ClassVar

from docutils import nodes
from docutils.parsers.rst.directives import choice, flag
from sphinx.directives.code import CodeBlock

from contourpy.util.mpl_renderer import MplRenderer


class PlotDirective(CodeBlock):
    has_content = True
    optional_arguments = 2

    option_spec: ClassVar[dict[str, Any]] = {  # type: ignore[misc]
        "separate-modes": flag,
        "source-position": lambda x: choice(x, ("below", "above", "none")),
    }

    # dict of string docname -> latest image index used.
    latest_image_index: ClassVar[dict[str, int]] = {}

    def _mpl_mode_header(self, mode: str) -> str:
        if mode == "light":
            return "import matplotlib.pyplot as plt;plt.style.use('default');\n"
        elif mode == "dark":
            return "import matplotlib as mpl;cycler=mpl.rcParams['axes.prop_cycle'];\n" \
                "import matplotlib.pyplot as plt;plt.style.use('dark_background');\n" \
                "mpl.rcParams['axes.prop_cycle']=cycler;\n"
        else:
            raise ValueError(f"Unexpected mode {mode}")

    def _temporary_show(self, renderer: MplRenderer, image_filenames: list[str]) -> None:
        # Temporary replacement for MplRenderer.show() to save to SVG file instead.
        docname = self.env.docname
        index = self.latest_image_index.get(docname, -1) + 1
        self.latest_image_index[docname] = index

        directory, filename = os.path.split(docname)
        output_directory = os.path.join(directory, "generated")
        if not os.path.isdir(output_directory):
            os.makedirs(output_directory)

        output_filename = f"{filename}_{index}.svg"

        renderer.save(os.path.join(output_directory, output_filename), transparent=True)
        image_filenames.append(os.path.join("generated", output_filename))

    def run(self: Any) -> list[Any]:
        source_position = self.options.get("source-position", "below")

        source = self.content
        combined_source = "\n".join(source)

        using_modes = "separate-modes" in self.options
        modes = ["light", "dark"] if using_modes else ["light"]

        svg_filenames: list[str] = []

        # Temporarily replace MplRenderer.show() to save to SVG file and include SVG files in
        # sphinx output. Should probably be in a context manager instead.
        old_show = getattr(MplRenderer, "show")
        setattr(MplRenderer, "show", lambda renderer: self._temporary_show(renderer, svg_filenames))
        for mode in modes:
            exec(self._mpl_mode_header(mode) + combined_source)
        setattr(MplRenderer, "show", old_show)

        images: list[nodes.Node] = []
        for i, svg_filename in enumerate(svg_filenames):
            image = nodes.image(uri=svg_filename)
            if using_modes:
                mode = modes[i % len(modes)]
                image["classes"].append(f"only-{mode}")
            container = nodes.container()
            container += image
            images += container

        if source_position == "none":
            return images
        else:
            code_block = super().run()
            if source_position == "above":
                return code_block + images
            else:
                return images + code_block


def setup(app: Any) -> dict[str, bool]:
    app.add_directive("plot", PlotDirective)
    return {"parallel_read_safe": True, "parallel_write_safe": True}



from __future__ import annotations

import numpy as np

from contourpy import _remove_z_mask, max_threads


def test_max_threads() -> None:
    n = max_threads()
    # Assume testing on machine with 2 or more cores.
    assert n > 1


def test_remove_z_mask() -> None:
    zlist = [[1.0, 2.0], [3.0, 4.0]]

    zz, mask = _remove_z_mask(zlist)
    assert isinstance(zz, np.ndarray)
    assert mask is None

    zarr = np.asarray(zlist)
    zz, mask = _remove_z_mask(zarr)
    assert isinstance(zz, np.ndarray)
    assert mask is None

    z = np.ma.array(zlist, mask=[[0, 0], [0, 0]])  # type: ignore[no-untyped-call]
    zz, mask = _remove_z_mask(zarr)
    assert isinstance(zz, np.ndarray)
    assert mask is None

    z = np.ma.array(zlist, mask=[[1, 0], [0, 0]])  # type: ignore[no-untyped-call]
    zz, mask = _remove_z_mask(z)
    assert isinstance(zz, np.ndarray)
    assert isinstance(mask, np.ndarray)
    assert mask.dtype == bool
    np.testing.assert_array_equal(mask, [[True, False], [False, False]])

    z = [[1.0, np.nan], [2.0, np.inf]]
    zz, mask = _remove_z_mask(z)
    assert isinstance(zz, np.ndarray)
    assert isinstance(mask, np.ndarray)
    assert mask.dtype == bool
    np.testing.assert_array_equal(mask, [[False, True], [False, True]])

    z = np.ma.array(zlist, mask=[[1, 0], [0, 0]])  # type: ignore[no-untyped-call]
    z[1][1] = np.nan
    zz, mask = _remove_z_mask(z)
    assert isinstance(zz, np.ndarray)
    assert isinstance(mask, np.ndarray)
    assert mask.dtype == bool
    np.testing.assert_array_equal(mask, [[True, False], [False, True]])



from __future__ import annotations

from contourpy import FillType, contour_generator

from .bench_base import BenchBase
from .util_bench import corner_mask_to_bool, corner_masks, datasets, fill_types, total_chunk_counts


class BenchFilledSerialChunk(BenchBase):
    params: tuple[list[str], list[str], list[FillType], list[str | bool], list[int], list[int]] = (
        ["serial"], datasets(), fill_types(), corner_masks(), [1000], total_chunk_counts(),
    )
    param_names: tuple[str, ...] = (
        "name", "dataset", "fill_type", "corner_mask", "n", "total_chunk_count",
    )

    def setup(
        self, name: str, dataset: str, fill_type: FillType, corner_mask: str | bool, n: int,
        total_chunk_count: int,
    ) -> None:
        self.set_xyz_and_levels(dataset, n, corner_mask != "no mask")

    def time_filled_serial_chunk(
        self, name: str, dataset: str, fill_type: FillType, corner_mask: str | bool, n: int,
        total_chunk_count: int,
    ) -> None:
        cont_gen = contour_generator(
            self.x, self.y, self.z, name=name, fill_type=fill_type,
            corner_mask=corner_mask_to_bool(corner_mask), total_chunk_count=total_chunk_count,
        )
        for i in range(len(self.levels)-1):
            cont_gen.filled(self.levels[i], self.levels[i+1])



from __future__ import annotations

import re
from typing import TYPE_CHECKING

import numpy as np
import pytest

from contourpy import FillType, LineType, contour_generator
from contourpy.util.data import random, simple

if TYPE_CHECKING:
    from _pytest._py.path import LocalPath


@pytest.mark.image
@pytest.mark.text
@pytest.mark.parametrize("show_text", [False, True])
@pytest.mark.parametrize("fill_type", FillType.__members__.values())
def test_debug_renderer_filled(show_text: bool, fill_type: FillType) -> None:
    from contourpy.util.mpl_renderer import MplDebugRenderer

    from .image_comparison import compare_images

    renderer = MplDebugRenderer(figsize=(4.5, 3), show_frame=show_text)
    x, y, z = simple((3, 4))
    cont_gen = contour_generator(x, y, z, fill_type=fill_type)
    levels = [0.25, 0.6]

    filled = cont_gen.filled(levels[0], levels[1])
    renderer.filled(filled, fill_type, color="C1")

    renderer.grid(x, y)
    if show_text:
        renderer.quad_numbers(x, y, z)
        renderer.z_levels(x, y, z, lower_level=levels[0], upper_level=levels[1])

    image_buffer = renderer.save_to_buffer()
    suffix = "" if show_text else "_no_text"
    compare_images(image_buffer, f"debug_renderer_filled{suffix}.png", f"{fill_type}")


@pytest.mark.image
@pytest.mark.text
@pytest.mark.parametrize("show_text", [False, True])
@pytest.mark.parametrize("line_type", LineType.__members__.values())
def test_debug_renderer_lines(show_text: bool, line_type: LineType) -> None:
    from contourpy.util.mpl_renderer import MplDebugRenderer

    from .image_comparison import compare_images

    renderer = MplDebugRenderer(figsize=(4.5, 3), show_frame=show_text)
    x, y, z = simple((3, 4))
    cont_gen = contour_generator(x, y, z, line_type=line_type)
    levels = [0.25, 0.6]

    for i, level in enumerate(levels):
        lines = cont_gen.lines(level)
        renderer.lines(lines, line_type, color=f"C{i}", linewidth=2)

    renderer.grid(x, y)
    if show_text:
        renderer.quad_numbers(x, y, z)
        renderer.point_numbers(x, y, z)

    image_buffer = renderer.save_to_buffer()
    suffix = "" if show_text else "_no_text"
    compare_images(image_buffer, f"debug_renderer_lines{suffix}.png", f"{line_type}")


@pytest.mark.image
@pytest.mark.text
@pytest.mark.parametrize("show_text", [False, True])
@pytest.mark.parametrize("fill_type", FillType.__members__.values())
def test_renderer_filled(show_text: bool, fill_type: FillType) -> None:
    from contourpy.util.mpl_renderer import MplRenderer

    from .image_comparison import compare_images

    renderer = MplRenderer(ncols=2, figsize=(8, 3), show_frame=False)
    x, y, z = random((3, 4), mask_fraction=0.35)
    for ax, quad_as_tri in enumerate((False, True)):
        cont_gen = contour_generator(x, y, z, fill_type=fill_type)

        levels = np.linspace(0.0, 1.0, 11)
        for i in range(len(levels)-1):
            filled = cont_gen.filled(levels[i], levels[i+1])
            if quad_as_tri:
                renderer.filled(filled, fill_type, ax=ax, color=f"C{i}", alpha=0.4)
            else:
                renderer.filled(filled, fill_type, ax=ax, color=f"C{i}")

        if quad_as_tri:
            renderer.grid(x, y, ax=ax, alpha=0.5, quad_as_tri_alpha=0.5)
            renderer.mask(x, y, z, ax=ax, color="red")
            if show_text:
                renderer.title("Title", ax=ax)
        else:
            renderer.grid(x, y, ax=ax, alpha=0.8, point_color="black")
            if show_text:
                renderer.title("Colored title", ax=ax, color="red")

    image_buffer = renderer.save_to_buffer()
    suffix = "" if show_text else "_no_text"
    compare_images(image_buffer, f"renderer_filled{suffix}.png", f"{fill_type}")


@pytest.mark.image
@pytest.mark.text
@pytest.mark.parametrize("show_text", [False, True])
@pytest.mark.parametrize("line_type", LineType.__members__.values())
def test_renderer_lines(show_text: bool, line_type: LineType) -> None:
    from contourpy.util.mpl_renderer import MplRenderer

    from .image_comparison import compare_images

    renderer = MplRenderer(ncols=2, figsize=(8, 3), show_frame=show_text)
    x, y, z = random((3, 4), mask_fraction=0.35)
    for ax, quad_as_tri in enumerate((False, True)):
        cont_gen = contour_generator(x, y, z, line_type=line_type)

        levels = np.linspace(0.1, 0.9, 9)
        for i in range(len(levels)):
            lines = cont_gen.lines(levels[i])
            if quad_as_tri:
                renderer.lines(lines, line_type, ax=ax, color=f"C{i}", alpha=0.5, linewidth=3)
            else:
                renderer.lines(lines, line_type, ax=ax, color=f"C{i}")

        if quad_as_tri:
            renderer.grid(x, y, ax=ax, alpha=0.2, quad_as_tri_alpha=0.2)
            renderer.mask(x, y, z, ax=ax, color="red")
            if show_text:
                renderer.z_values(x, y, z, ax=ax, fmt=".2f", quad_as_tri=True)
                renderer.title("Title", ax=ax)
        else:
            renderer.grid(x, y, ax=ax, point_color="black")
            if show_text:
                renderer.z_values(x, y, z, ax=ax, fmt=".2f", color="blue")
                renderer.title("Colored title", ax=ax, color="red")

    image_buffer = renderer.save_to_buffer()
    suffix = "" if show_text else "_no_text"
    compare_images(
        image_buffer, f"renderer_lines{suffix}.png", f"{line_type}",
    )


@pytest.mark.image
@pytest.mark.parametrize("transparent", [False, True])
def test_save_png(transparent: bool, tmpdir: LocalPath) -> None:
    from PIL import Image

    from contourpy.util.mpl_renderer import MplRenderer

    renderer = MplRenderer(figsize=(4, 3))
    filename = (tmpdir / "mpl.png").strpath
    renderer.save(filename, transparent)

    # Testing that a PNG file is produced of the correct size and format.
    # Not testing the actual image produced except for the first pixel to confirm transparency.
    with Image.open(filename) as image:
        assert image.format == "PNG"
        assert image.mode == "RGBA"
        assert image.size == (400, 300)
        rgba = image.getpixel((0, 0))
        assert rgba[:3] == (255, 255, 255)
        assert rgba[3] == 0 if transparent else 255


@pytest.mark.image
@pytest.mark.parametrize("transparent", [False, True])
def test_save_svg(transparent: bool, tmpdir: LocalPath) -> None:
    from contourpy.util.mpl_renderer import MplRenderer

    renderer = MplRenderer(figsize=(4, 3))
    filename = (tmpdir / "mpl.svg").strpath
    renderer.save(filename, transparent)

    # Rather simplistic check of SVG file contents.
    with open(filename) as f:
        svg = f.read()

    assert len(re.findall("<svg ", svg)) == 1
    count0 = len(re.findall("fill: none", svg))
    count1 = len(re.findall("fill: #ffffff", svg))
    if transparent:
        assert count0 == 6 and count1 == 0
    else:
        assert count0 == 4 and count1 == 2



from __future__ import annotations

import math
from typing import TYPE_CHECKING

import numpy as np
import pytest

from contourpy import ContourGenerator, FillType, LineType, ZInterp, contour_generator, max_threads

from . import util_test

if TYPE_CHECKING:
    from numpy.typing import ArrayLike

    from contourpy._contourpy import CoordinateArray


@pytest.fixture
def xyz_3x3_as_lists() -> tuple[list[list[int]], ...]:
    x = [[0, 1, 2], [0, 1, 2], [0, 1, 2]]
    y = [[0, 0, 0], [1, 1, 1], [2, 2, 2]]
    z = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    return x, y, z


@pytest.fixture
def xyz_2x3_as_lists() -> tuple[list[list[int]], ...]:  # shape (2, 3)
    x = [[0, 1, 2], [0, 1, 2]]
    y = [[0, 0, 0], [1, 1, 1]]
    z = [[0, 1, 2], [3, 4, 5]]
    return x, y, z


@pytest.fixture
def xyz_7x5_as_arrays() -> tuple[CoordinateArray, ...]:  # shape (7, 5)
    x, y = np.meshgrid([0, 1, 2, 3, 4], [0, 1, 2, 3, 4, 5, 6])
    z = x + y
    return x, y, z


@pytest.mark.parametrize("name", util_test.all_names())
@pytest.mark.parametrize("z", ([1], [[[1]]]))
def test_ndim_z(
    xyz_3x3_as_lists: tuple[list[list[int]], ...],
    name: str,
    z: ArrayLike,
) -> None:
    x, y, _ = xyz_3x3_as_lists
    with pytest.raises(TypeError):
        contour_generator(x, y, z, name=name)


@pytest.mark.parametrize("name", util_test.all_names())
@pytest.mark.parametrize("all_xyz", ([[1]], [[1], [2]], [[1, 2]]))
def test_z_shape_too_small(all_xyz: ArrayLike, name: str) -> None:
    with pytest.raises(TypeError):
        contour_generator(z=all_xyz, name=name)


@pytest.mark.parametrize("name", util_test.all_names())
@pytest.mark.parametrize("wrong_ndim", (None, [1], [[[1]]]))
def test_diff_ndim_xy(
    xyz_3x3_as_lists: tuple[list[list[int]], ...],
    name: str,
    wrong_ndim: ArrayLike,
) -> None:
    x, y, z = xyz_3x3_as_lists
    with pytest.raises(TypeError):
        contour_generator(wrong_ndim, y, z, name=name)
    with pytest.raises(TypeError):
        contour_generator(x, wrong_ndim, z, name=name)


@pytest.mark.parametrize("name", util_test.all_names())
def test_xy_None(xyz_3x3_as_lists: tuple[list[list[int]], ...], name: str) -> None:
    _, _, z = xyz_3x3_as_lists
    contour_generator(None, None, z, name=name)


@pytest.mark.parametrize("name", util_test.all_names())
def test_xy_not_specified(xyz_3x3_as_lists: tuple[list[list[int]], ...], name: str) -> None:
    _, _, z = xyz_3x3_as_lists
    contour_generator(z=z, name=name)


@pytest.mark.parametrize("name", util_test.all_names())
def test_xy_1d(name: str) -> None:
    z = [[0, 1, 2], [3, 4, 5]]
    contour_generator([0, 1, 2], [0, 1], z, name=name)
    with pytest.raises(TypeError):
        contour_generator([0, 1], [0, 1], z, name=name)
    with pytest.raises(TypeError):
        contour_generator([0, 1, 2, 3], [0, 1], z, name=name)
    with pytest.raises(TypeError):
        contour_generator([0, 1, 2], [0], z, name=name)
    with pytest.raises(TypeError):
        contour_generator([0, 1, 2], [0, 1, 2], z, name=name)


@pytest.mark.parametrize("name", util_test.all_names())
def test_xy_ndim_more_than_2(name: str) -> None:
    z = [[0, 1, 2], [3, 4, 5]]
    msg = "Inputs x and y must be None, 1D or 2D, not 3D"
    with pytest.raises(TypeError, match=msg):
        contour_generator([[[1]]], [[[2]]], z, name=name)


@pytest.mark.parametrize("name", util_test.all_names())
@pytest.mark.parametrize("diff_shape", ([[1, 2, 3, 4], [5, 6, 7, 8]], [[1, 2], [3, 4], [5, 6]]))
def test_xyz_diff_shapes(
    xyz_3x3_as_lists: tuple[list[list[int]], ...],
    name: str,
    diff_shape: ArrayLike,
) -> None:
    x, y, z = xyz_3x3_as_lists
    with pytest.raises(TypeError):
        contour_generator(diff_shape, y, z, name=name)
    with pytest.raises(TypeError):
        contour_generator(x, diff_shape, z, name=name)
    with pytest.raises(TypeError):
        contour_generator(x, y, diff_shape, name=name)


@pytest.mark.parametrize("name", util_test.corner_mask_names())
def test_corner_mask(xyz_3x3_as_lists: tuple[list[list[int]], ...], name: str) -> None:
    x, y, z = xyz_3x3_as_lists
    for corner_mask in [False, True]:
        cont_gen = contour_generator(x, y, z, name=name, corner_mask=corner_mask)
        assert cont_gen.corner_mask == corner_mask


def test_corner_mask_not_supported(xyz_3x3_as_lists: tuple[list[list[int]], ...]) -> None:
    name = "mpl2005"
    x, y, z = xyz_3x3_as_lists
    msg = f"{name} contour generator does not support corner_mask=True"
    with pytest.raises(ValueError, match=msg):
        contour_generator(x, y, z, name=name, corner_mask=True)


@pytest.mark.parametrize("name", util_test.all_names())
def test_chunk_size_negative(xyz_3x3_as_lists: tuple[list[list[int]], ...], name: str) -> None:
    x, y, z = xyz_3x3_as_lists
    msg = "chunk_size cannot be negative"
    with pytest.raises(ValueError, match=msg):
        contour_generator(x, y, z, name=name, chunk_size=-1)
    with pytest.raises(ValueError, match=msg):
        contour_generator(x, y, z, name=name, chunk_size=(-1, 0))
    with pytest.raises(ValueError, match=msg):
        contour_generator(x, y, z, name=name, chunk_size=(0, -1))


@pytest.mark.parametrize("name", util_test.all_names())
@pytest.mark.parametrize("chunk_size", np.arange(9))
def test_chunk_size_1d(
    xyz_7x5_as_arrays: tuple[CoordinateArray, ...],
    name: str,
    chunk_size: int,
) -> None:
    x, y, z = xyz_7x5_as_arrays
    ny, nx = z.shape
    cont_gen = contour_generator(x, y, z, name=name, chunk_size=chunk_size)
    ret_y_chunk_size, ret_x_chunk_size = cont_gen.chunk_size
    ret_y_chunk_count, ret_x_chunk_count = cont_gen.chunk_count
    if chunk_size == 0:
        assert ret_x_chunk_size == nx-1
        assert ret_y_chunk_size == ny-1
    else:
        assert ret_x_chunk_size == min(chunk_size, nx-1)
        assert ret_y_chunk_size == min(chunk_size, ny-1)
    assert ret_y_chunk_count*ret_y_chunk_size >= ny-1
    assert (ret_y_chunk_count-1)*ret_y_chunk_size < ny-1
    assert ret_x_chunk_count*ret_x_chunk_size >= nx-1
    assert (ret_x_chunk_count-1)*ret_x_chunk_size < nx-1


@pytest.mark.parametrize("name", util_test.all_names())
@pytest.mark.parametrize("x_chunk_size", np.arange(9))
@pytest.mark.parametrize("y_chunk_size", np.arange(9))
def test_chunk_size_2d(
    xyz_7x5_as_arrays: tuple[CoordinateArray, ...],
    name: str,
    x_chunk_size: int,
    y_chunk_size: int,
) -> None:
    x, y, z = xyz_7x5_as_arrays
    ny, nx = z.shape
    cont_gen = contour_generator(x, y, z, name=name, chunk_size=(y_chunk_size, x_chunk_size))
    ret_y_chunk_size, ret_x_chunk_size = cont_gen.chunk_size
    ret_y_chunk_count, ret_x_chunk_count = cont_gen.chunk_count
    if x_chunk_size == 0:
        assert ret_x_chunk_size == nx-1
    else:
        assert ret_x_chunk_size == min(x_chunk_size, nx-1)
    if y_chunk_size == 0:
        assert ret_y_chunk_size == ny-1
    else:
        assert ret_y_chunk_size == min(y_chunk_size, ny-1)
    assert ret_y_chunk_count*ret_y_chunk_size >= ny-1
    assert (ret_y_chunk_count-1)*ret_y_chunk_size < ny-1
    assert ret_x_chunk_count*ret_x_chunk_size >= nx-1
    assert (ret_x_chunk_count-1)*ret_x_chunk_size < nx-1


def test_chunk_size_and_count(xyz_7x5_as_arrays: tuple[CoordinateArray, ...]) -> None:
    x, y, z = xyz_7x5_as_arrays
    name = "serial"
    msg = "Only one of chunk_size, chunk_count and total_chunk_count should be set"
    with pytest.raises(ValueError, match=msg):
        contour_generator(x, y, z, name=name, chunk_size=1, chunk_count=1)
    with pytest.raises(ValueError, match=msg):
        contour_generator(x, y, z, name=name, chunk_size=1, total_chunk_count=1)
    with pytest.raises(ValueError, match=msg):
        contour_generator(x, y, z, name=name, chunk_count=1, total_chunk_count=1)


@pytest.mark.parametrize("name", util_test.all_names())
@pytest.mark.parametrize(
    "chunk_count, ret_chunk_count",
    [[0, (1, 1)], [1, (1, 1)], [2, (2, 2)], [3, (3, 2)], [4, (3, 4)], [9, (6, 4)]],
)
def test_chunk_count_1d(
    xyz_7x5_as_arrays: tuple[CoordinateArray, ...],
    name: str,
    chunk_count: int,
    ret_chunk_count: tuple[int, int],
) -> None:
    x, y, z = xyz_7x5_as_arrays
    ny, nx = z.shape
    cont_gen = contour_generator(x, y, z, name=name, chunk_count=chunk_count)
    ret_y_chunk_size, ret_x_chunk_size = cont_gen.chunk_size
    assert cont_gen.chunk_count == ret_chunk_count
    ret_y_chunk_count, ret_x_chunk_count = ret_chunk_count
    assert ret_y_chunk_count*ret_y_chunk_size >= ny-1
    assert (ret_y_chunk_count-1)*ret_y_chunk_size < ny-1
    assert ret_x_chunk_count*ret_x_chunk_size >= nx-1
    assert (ret_x_chunk_count-1)*ret_x_chunk_size < nx-1


@pytest.mark.parametrize("name", util_test.all_names())
@pytest.mark.parametrize(
    "chunk_count, ret_chunk_count",
    [[(0, 1), (1, 1)], [(1, 0), (1, 1)], [(2, 3), (2, 2)], [(3, 2), (3, 2)], [(1, 9), (1, 4)],
     [(9, 1), (6, 1)]],
)
def test_chunk_count_2d(
    xyz_7x5_as_arrays: tuple[CoordinateArray, ...],
    name: str,
    chunk_count: tuple[int, int],
    ret_chunk_count: tuple[int, int],
) -> None:
    x, y, z = xyz_7x5_as_arrays
    ny, nx = z.shape
    cont_gen = contour_generator(x, y, z, name=name, chunk_count=chunk_count)
    ret_y_chunk_size, ret_x_chunk_size = cont_gen.chunk_size
    assert cont_gen.chunk_count == ret_chunk_count
    ret_y_chunk_count, ret_x_chunk_count = ret_chunk_count
    assert ret_y_chunk_count*ret_y_chunk_size >= ny-1
    assert (ret_y_chunk_count-1)*ret_y_chunk_size < ny-1
    assert ret_x_chunk_count*ret_x_chunk_size >= nx-1
    assert (ret_x_chunk_count-1)*ret_x_chunk_size < nx-1


@pytest.mark.parametrize("name", util_test.all_names())
@pytest.mark.parametrize(
    "total_chunk_count, ret_chunk_count",
    [[0, (1, 1)], [1, (1, 1)], [2, (2, 1)], [3, (3, 1)], [4, (2, 2)], [6, (3, 2)], [9, (3, 2)],
     [25, (6, 4)]],
)
def test_total_chunk_count(
    xyz_7x5_as_arrays: tuple[CoordinateArray, ...],
    name: str,
    total_chunk_count: int,
    ret_chunk_count: tuple[int, int],
) -> None:
    x, y, z = xyz_7x5_as_arrays
    ny, nx = z.shape
    cont_gen = contour_generator(x, y, z, name=name, total_chunk_count=total_chunk_count)
    ret_y_chunk_size, ret_x_chunk_size = cont_gen.chunk_size
    assert cont_gen.chunk_count == ret_chunk_count
    ret_y_chunk_count, ret_x_chunk_count = ret_chunk_count
    assert ret_y_chunk_count*ret_y_chunk_size >= ny-1
    assert (ret_y_chunk_count-1)*ret_y_chunk_size < ny-1
    assert ret_x_chunk_count*ret_x_chunk_size >= nx-1
    assert (ret_x_chunk_count-1)*ret_x_chunk_size < nx-1


def test_name_invalid(xyz_3x3_as_lists: tuple[list[list[int]], ...]) -> None:
    x, y, z = xyz_3x3_as_lists
    name = "some invalid name"
    msg = f"Unrecognised contour generator name: {name}"
    with pytest.raises(ValueError, match=msg):
        _ = contour_generator(x, y, z, name=name)


@pytest.mark.parametrize("name", ["mpl2005", "mpl2014"])
@pytest.mark.parametrize(
    "line_type", [LineType.Separate, LineType.ChunkCombinedCode, LineType.ChunkCombinedOffset])
def test_line_type_not_supported(
    xyz_3x3_as_lists: tuple[list[list[int]], ...],
    name: str,
    line_type: LineType,
) -> None:
    x, y, z = xyz_3x3_as_lists
    msg = f"{name} contour generator does not support line_type {line_type}"
    with pytest.raises(ValueError, match=msg):
        contour_generator(x, y, z, name=name, line_type=line_type)


@pytest.mark.parametrize("name", ["mpl2005", "mpl2014"])
@pytest.mark.parametrize("fill_type", [
    FillType.OuterOffset, FillType.ChunkCombinedCode, FillType.ChunkCombinedOffset,
    FillType.ChunkCombinedCodeOffset, FillType.ChunkCombinedOffsetOffset])
def test_fill_type_not_supported(
    xyz_3x3_as_lists: tuple[list[list[int]], ...],
    name: str,
    fill_type: FillType,
) -> None:
    x, y, z = xyz_3x3_as_lists
    msg = f"{name} contour generator does not support fill_type {fill_type}"
    with pytest.raises(ValueError, match=msg):
        contour_generator(x, y, z, name=name, fill_type=fill_type)


@pytest.mark.parametrize("name", util_test.all_names())
def test_properties(xyz_3x3_as_lists: tuple[list[list[int]], ...], name: str) -> None:
    # Test that each ContourGenerator class implements all the required properties.
    x, y, z = xyz_3x3_as_lists
    cont_gen = contour_generator(x, y, z, name=name)
    _, _ = cont_gen.chunk_count
    _, _ = cont_gen.chunk_size
    _ = cont_gen.corner_mask
    _ = cont_gen.fill_type
    _ = cont_gen.line_type
    _ = cont_gen.quad_as_tri
    _ = cont_gen.thread_count
    _ = cont_gen.z_interp


@pytest.mark.parametrize("name", util_test.quad_as_tri_names())
def test_quad_as_tri(xyz_3x3_as_lists: tuple[list[list[int]], ...], name: str) -> None:
    x, y, z = xyz_3x3_as_lists
    for quad_as_tri in [False, True]:
        cont_gen = contour_generator(x, y, z, name=name, quad_as_tri=quad_as_tri)
        assert cont_gen.quad_as_tri == quad_as_tri


def test_quad_as_tri_not_supported(xyz_3x3_as_lists: tuple[list[list[int]], ...]) -> None:
    x, y, z = xyz_3x3_as_lists
    name = "mpl2005"
    msg = f"{name} contour generator does not support quad_as_tri=True"
    with pytest.raises(ValueError, match=msg):
        contour_generator(x, y, z, name=name, quad_as_tri=True)


@pytest.mark.parametrize("chunk_size", [0, 1, 2])
@pytest.mark.parametrize("thread_count", [0, 1, 2])
def test_thread_count(
    xyz_7x5_as_arrays: tuple[CoordinateArray, ...],
    chunk_size: int,
    thread_count: int,
) -> None:
    name = "threaded"
    x, y, z = xyz_7x5_as_arrays
    cont_gen = contour_generator(
        x, y, z, name=name, chunk_size=chunk_size, thread_count=thread_count)
    ret_thread_count = cont_gen.thread_count
    ret_chunk_count = math.prod(cont_gen.chunk_count)
    max_thread_count = max_threads()
    if chunk_size == 0:
        assert ret_chunk_count == 1
        assert ret_thread_count == 1
    elif thread_count == 0:
        assert ret_thread_count == min(max_thread_count, ret_chunk_count)
    else:
        assert ret_thread_count == min(max_thread_count, ret_chunk_count, ret_thread_count)


@pytest.mark.parametrize("name", util_test.all_names(exclude="threaded"))
def test_thread_count_not_supported(
    xyz_3x3_as_lists: tuple[list[list[int]], ...],
    name: str,
) -> None:
    x, y, z = xyz_3x3_as_lists
    thread_count = 2
    msg = f"{name} contour generator does not support thread_count {thread_count}"
    with pytest.raises(ValueError, match=msg):
        contour_generator(x, y, z, name=name, thread_count=thread_count)


def test_enums_as_strings(xyz_3x3_as_lists: tuple[list[list[int]], ...]) -> None:
    x, y, z = xyz_3x3_as_lists
    cg = contour_generator(
        x, y, z, line_type="SeparateCode", fill_type="ChunkCombinedCodeOffset", z_interp="Log")
    assert cg.line_type == LineType.SeparateCode
    assert cg.fill_type == FillType.ChunkCombinedCodeOffset
    assert cg.z_interp == ZInterp.Log


@pytest.mark.parametrize("name", util_test.all_names())
def test_is_contour_generator(name: str, xyz_3x3_as_lists: tuple[list[list[int]], ...]) -> None:
    x, y, z = xyz_3x3_as_lists
    cg = contour_generator(x, y, z, name=name)
    assert isinstance(cg, ContourGenerator)


@pytest.mark.parametrize("name", util_test.all_names())
def test_z_interp_none_to_linear(name: str, xyz_3x3_as_lists: tuple[list[list[int]], ...]) -> None:
    x, y, z = xyz_3x3_as_lists
    cg = contour_generator(x, y, z, name=name, z_interp=None)
    assert cg.z_interp == ZInterp.Linear


@pytest.mark.parametrize("name", ["mpl2005", "mpl2014"])
def test_z_interp_not_supported(name: str, xyz_3x3_as_lists: tuple[list[list[int]], ...]) -> None:
    x, y, z = xyz_3x3_as_lists
    msg = f"{name} contour generator does not support z_interp ZInterp.Log"
    with pytest.raises(ValueError, match=msg):
        contour_generator(x, y, z, name=name, z_interp=ZInterp.Log)



from __future__ import annotations

from contourpy import FillType, LineType, max_threads


def corner_mask_to_bool(corner_mask: str | bool) -> bool:
    if isinstance(corner_mask, bool):
        return corner_mask
    else:
        return False


def corner_masks() -> list[str | bool]:
    return ["no mask", False, True]


def datasets() -> list[str]:
    return ["simple", "random"]


def fill_types() -> list[FillType]:
    return list(FillType.__members__.values())


def line_types() -> list[LineType]:
    return list(LineType.__members__.values())


def problem_sizes() -> list[int]:
    return [10, 30, 100, 300, 1000]


def thread_counts() -> list[int]:
    thread_counts = [1, 2, 4, 6, 8]
    return list(filter(lambda n: n <= max(max_threads(), 1), thread_counts))


def total_chunk_counts() -> list[int]:
    return [4, 12, 40, 120]



from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING, Iterator

import numpy as np
import pytest

from contourpy import FillType, LineType, contour_generator
from contourpy.util.data import random

if TYPE_CHECKING:
    from _pytest._py.path import LocalPath
    from _pytest.logging import LogCaptureFixture
    from selenium.webdriver.remote.webdriver import WebDriver

bokeh_renderer = pytest.importorskip("contourpy.util.bokeh_renderer")


@pytest.fixture(scope="session")
def driver(driver_path: str) -> Iterator[WebDriver]:
    # Based on Bokeh's tests/support/plugins/selenium.py
    def chrome() -> WebDriver:
        from selenium.webdriver.chrome.options import Options
        from selenium.webdriver.chrome.webdriver import WebDriver as Chrome

        options = Options()
        options.add_argument("--headless")  # type: ignore[no-untyped-call]
        options.add_argument("--no-sandbox")  # type: ignore[no-untyped-call]

        if driver_path:
            from selenium.webdriver.chrome.service import Service
            service = Service(executable_path=driver_path)
            return Chrome(options=options, service=service)
        else:
            return Chrome(options=options)

    driver = chrome()
    driver.implicitly_wait(10)
    yield driver
    driver.quit()


def test_chrome_version(driver: WebDriver) -> None:
    # Print version of chrome used for export to PNG by Bokeh as test images, particularly those
    # containing text, are sensitive to chrome version.
    capabilities = driver.capabilities
    print("Browser used by Bokeh:", capabilities["browserName"], capabilities["browserVersion"])
    if "chrome" in capabilities:
        print("Chromedriver:", capabilities["chrome"]["chromedriverVersion"])


@pytest.mark.image
@pytest.mark.text
@pytest.mark.parametrize("show_text", [False, True])
@pytest.mark.parametrize("fill_type", FillType.__members__.values())
def test_renderer_filled_bokeh(show_text: bool, fill_type: FillType, driver: WebDriver) -> None:
    from .image_comparison import compare_images

    renderer = bokeh_renderer.BokehRenderer(ncols=2, figsize=(8, 3), show_frame=False)
    x, y, z = random((3, 4), mask_fraction=0.35)
    for ax, quad_as_tri in enumerate((False, True)):
        cont_gen = contour_generator(x, y, z, fill_type=fill_type)

        levels = np.linspace(0.0, 1.0, 11)
        for i in range(len(levels)-1):
            filled = cont_gen.filled(levels[i], levels[i+1])
            if quad_as_tri:
                renderer.filled(filled, fill_type, ax=ax, color=f"C{i}", alpha=0.4)
            else:
                renderer.filled(filled, fill_type, ax=ax, color=f"C{i}")

        if quad_as_tri:
            renderer.grid(x, y, ax=ax, alpha=0.5, quad_as_tri_alpha=0.5)
            renderer.mask(x, y, z, ax=ax, color="red")
            if show_text:
                renderer.title("Title", ax=ax)
        else:
            renderer.grid(x, y, ax=ax, alpha=0.8, point_color="black")
            if show_text:
                renderer.title("Colored title", ax=ax, color="red")

    image_buffer = renderer.save_to_buffer(webdriver=driver)
    suffix = "" if show_text else "_no_text"
    compare_images(
        image_buffer, f"bokeh_renderer_filled{suffix}.png", f"{fill_type}", mean_threshold=0.03,
    )


@pytest.mark.image
@pytest.mark.text
@pytest.mark.parametrize("show_text", [False, True])
@pytest.mark.parametrize("line_type", LineType.__members__.values())
def test_renderer_lines_bokeh(show_text: bool, line_type: LineType, driver: WebDriver) -> None:
    from .image_comparison import compare_images

    renderer = bokeh_renderer.BokehRenderer(ncols=2, figsize=(8, 3), show_frame=show_text)
    x, y, z = random((3, 4), mask_fraction=0.35)
    for ax, quad_as_tri in enumerate((False, True)):
        cont_gen = contour_generator(x, y, z, line_type=line_type)

        levels = np.linspace(0.1, 0.9, 9)
        for i in range(len(levels)):
            lines = cont_gen.lines(levels[i])
            if quad_as_tri:
                renderer.lines(lines, line_type, ax=ax, color=f"C{i}", alpha=0.5, linewidth=3)
            else:
                renderer.lines(lines, line_type, ax=ax, color=f"C{i}")

        if quad_as_tri:
            renderer.grid(x, y, ax=ax, alpha=0.2, quad_as_tri_alpha=0.2)
            renderer.mask(x, y, z, ax=ax, color="red")
            if show_text:
                renderer.z_values(x, y, z, ax=ax, fmt=".2f", quad_as_tri=True)
                renderer.title("Title", ax=ax)
        else:
            renderer.grid(x, y, ax=ax, point_color="black")
            if show_text:
                renderer.z_values(x, y, z, ax=ax, fmt=".2f", color="blue")
                renderer.title("Colored title", ax=ax, color="red")

    image_buffer = renderer.save_to_buffer(webdriver=driver)
    suffix = "" if show_text else "_no_text"
    compare_images(
        image_buffer, f"bokeh_renderer_lines{suffix}.png", f"{line_type}", mean_threshold=0.03,
    )


@pytest.mark.image
@pytest.mark.parametrize("transparent", [False, True])
def test_save_png(
    transparent: bool, tmpdir: LocalPath, driver: WebDriver, caplog: LogCaptureFixture,
) -> None:
    from PIL import Image

    renderer = bokeh_renderer.BokehRenderer(figsize=(4, 3), show_frame=False)
    filename = (tmpdir / "bokeh.png").strpath

    with caplog.at_level(logging.ERROR):  # Hide Bokeh warnings about no renderers.
        renderer.save(filename, transparent, webdriver=driver)

    # Testing that a PNG file is produced of the correct size and format.
    # Not testing the actual image produced except for the first pixel to confirm transparency.
    with Image.open(filename) as image:
        assert image.format == "PNG"
        assert image.mode == "RGBA"
        assert image.size == (400, 300)
        rgba = image.getpixel((0, 0))

        # Transparent background not working for PNG export.
        # assert rgba[3] == 0 if transparent else 255
        assert rgba[3] == 255


@pytest.mark.image
@pytest.mark.parametrize("transparent", [False, True])
def test_save_svg(
    transparent: bool, tmpdir: LocalPath, driver: WebDriver, caplog: LogCaptureFixture,
) -> None:
    renderer = bokeh_renderer.BokehRenderer(figsize=(4, 3), show_frame=False, want_svg=True)
    filename = (tmpdir / "bokeh.svg").strpath

    with caplog.at_level(logging.ERROR):  # Hide Bokeh warnings about no renderers.
        renderer.save(filename, transparent, webdriver=driver)

    # Rather simplistic check of SVG file contents.
    with open(filename) as f:
        svg = f.read()

    assert svg[:5] == "<svg "
    count0 = len(re.findall('fill-opacity="0"', svg))
    count1 = len(re.findall('fill-opacity="1"', svg))
    if transparent:
        assert count0 == 2 and count1 == 0
    else:
        assert count0 == 2 and count1 == 2



from __future__ import annotations

from typing import Callable

import pytest

from contourpy import FillType, LineType, ZInterp
from contourpy.enum_util import as_fill_type, as_line_type, as_z_interp

from . import util_test


@pytest.mark.parametrize("name, value", util_test.all_fill_types_str_value())
def test_fill_type(name: str, value: int) -> None:
    t = FillType.__members__[name]
    assert t.name == name
    assert t.value == value


@pytest.mark.parametrize("name, value", util_test.all_line_types_str_value())
def test_line_type(name: str, value: int) -> None:
    t = LineType.__members__[name]
    assert t.name == name
    assert t.value == value


def test_all_fill_types() -> None:
    # Check that all_fill_types() matches FillType.__members__
    fill_types = dict(util_test.all_fill_types_str_value())
    for name, enum in dict(FillType.__members__).items():
        assert name in fill_types
        assert fill_types[name] == enum.value


def test_all_line_types() -> None:
    # Check that all_line_types() matches LineType.__members__
    line_types = dict(util_test.all_line_types_str_value())
    for name, enum in dict(LineType.__members__).items():
        assert name in line_types
        assert line_types[name] == enum.value


def test_all_z_interps() -> None:
    # Check that all_z_interps() matches ZInterp.__members__
    z_interps = dict(util_test.all_z_interps_str_value())
    for name, enum in dict(ZInterp.__members__).items():
        assert name in z_interps
        assert z_interps[name] == enum.value


@pytest.mark.parametrize(
    ["enum_type", "from_string_function"],
    [(FillType, as_fill_type), (LineType, as_line_type), (ZInterp, as_z_interp)])
def test_string_to_enum(
    enum_type: FillType | LineType | ZInterp,
    from_string_function: Callable[[FillType | LineType | ZInterp | str],
                                   FillType | LineType | ZInterp],
) -> None:
    for name, enum in enum_type.__members__.items():
        line_type = from_string_function(name)
        assert line_type == enum

    msg = "'unknown'"
    with pytest.raises(KeyError, match=msg):
        _ = from_string_function("unknown")



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import numpy as np

from contourpy.util.data import random, simple

if TYPE_CHECKING:
    import numpy.typing as npt


class BenchBase:
    levels: npt.NDArray[np.float64]
    timeout: int = 120  # Some rendering benchmarks can take more than the default minute.
    x: npt.NDArray[np.float64]
    y: npt.NDArray[np.float64]
    z: npt.NDArray[np.float64] | np.ma.MaskedArray[Any, Any]

    def set_xyz_and_levels(self, dataset: str, n: int, want_mask: bool) -> None:
        if dataset == "random":
            mask_fraction = 0.05 if want_mask else 0.0
            self.x, self.y, self.z = random((n, n), mask_fraction=mask_fraction)
            self.levels = np.arange(0.0, 1.01, 0.1)
        elif dataset == "simple":
            self.x, self.y, self.z = simple((n, n), want_mask=want_mask)
            self.levels = np.arange(-1.0, 1.01, 0.1)
        else:
            raise NotImplementedError()
